from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Tuple, Type
import typing

from marshmallow import Schema
from marshmallow.fields import Field
import marshmallow_dataclass

from allophant.config import MultiheadAttentionConfig, ProjectionEntryConfig


class DependencyCycleError(Exception):
    """Raised when a dependency cycle is detected"""


@dataclass
class AttributeNode:
    """
    Contains the name, number of variants and dependencies of a phonetic attribute node

    :param name: Name of the attribute
    :param size: Number of variants of the attribute
    :param dependencies: Attribute names of each dependency
    """

    name: str
    size: int
    time_layer_config: Optional[MultiheadAttentionConfig] = None
    dependencies: List[str] = field(default_factory=list)

    def with_offset(self, offset: int = 1):
        """
        Creates a new phonetic attribute node with an offset added to the size
        but the same name and dependencies

        :param offset: An offset that is added to the node size

        :return: A new phonetic attribute node derived from the current node with `offset` added to its size
        """
        return AttributeNode(self.name, self.size + offset, self.time_layer_config, self.dependencies)


class AttributeGraph:
    """
    A graph of phonetic attributes which also stores the number of variants of
    each attribute and provides methods for topological sorting
    """

    _nodes: List[AttributeNode]
    _node_indices: Dict[str, int]
    _edges: List[List[int]]

    def __init__(self, nodes: Iterable[AttributeNode]) -> None:
        """
        Builds an attribute graph from a sequence of attributes with their number of variants and dependencies

        :param nodes: An iterable of attribute, size and dependencies tuples
        """
        self._nodes = []
        self._node_indices = {}

        for index, node in enumerate(nodes):
            self._nodes.append(node)
            self._node_indices[node.name] = index

        self._edges = [
            [
                self._node_indices[dependency]
                for dependency in node.dependencies
                if not ProjectionEntryConfig.OUTPUT_PATTERN.match(dependency)
            ]
            for node in self._nodes
        ]

    def sizes(self) -> Iterator[int]:
        """
        Generates the number of variants for each node in the graph

        :return: A generator over the number of variants for each node
        """
        return (node.size for node in self._nodes)

    def names(self) -> Iterator[str]:
        """
        Generates the phonetic attribute names for each node in the graph

        :return: A generator over the phonetic attribute names for each node
        """
        return (node.name for node in self._nodes)

    @property
    def nodes(self) -> List[AttributeNode]:
        return self._nodes

    def get(self, node: str | int) -> AttributeNode | None:
        """
        Retrieves a phonetic attribute node by either its name or its index in the graph

        :param node: The name or index of a phonetic attribute node

        :return: The attribute node with name or index `node` or `None` if a
            name was given but not found in the graph
        """
        if isinstance(node, str):
            node_index = self._node_indices.get(node)
            if node_index is None:
                return None
            node = node_index

        return self._nodes[node]

    def __getitem__(self, node: str | int) -> AttributeNode:
        if isinstance(node, str):
            node = self._node_indices[node]
        return self._nodes[node]

    def __contains__(self, node_name: str) -> bool:
        return node_name in self._node_indices

    def __len__(self) -> int:
        return len(self._nodes)

    def strongly_connected_components(self) -> Iterator[List[AttributeNode]]:
        """
        Yields the strongly connected components (SCC) of the graph in reverse topological order using Tarjan's SCC algorithm.
        Components with more than one element indicate a cycle and contain every node that participates in it.

        :return: A generator over all strongly connected components of the graph, represented by lists of :py:class:`AttributeNode` instances
        """
        # Immediately return on empty node
        if not self._nodes:
            return

        stack: List[int] = []
        call_stack: List[Tuple[int, int, int]] = [(0, 0, 0)]
        visited_count = 0
        low_map = {}
        remaining_nodes = iter(range(1, len(self)))

        while True:
            if call_stack:
                node, index, lowlink = call_stack.pop()
            else:
                # If the call stack is empty start the next traversal in case
                # the graph is disconnected
                try:
                    node = next(remaining_nodes)
                except StopIteration:
                    # End traversal if there are no remaining nodes
                    return
                index = 0
                lowlink = visited_count

            if index == 0:
                # Handle the first edge of a node if it wasn't already visited
                if node in low_map:
                    continue
                low_map[node] = lowlink
                visited_count += 1
                stack.append(node)
            else:
                low_map[node] = min(low_map[node], low_map[self._edges[node][index - 1]])

            if index < len(self._edges[node]):
                # If there are more outgoing edges from the current node, first
                # visit the next edge and then the target of the current edge
                call_stack.extend(
                    (
                        (node, index + 1, lowlink),
                        (self._edges[node][index], 0, visited_count),
                    )
                )
                continue

            # If a component is connected, reconstruct and return the current
            # strongly connected component by backtracking
            if lowlink == low_map[node]:
                components = []
                while stack:
                    stack_node = stack.pop()
                    components.append(stack_node)
                    low_map[stack_node] = len(self)
                    if stack_node == node:
                        break

                # If this list has a length of one, the containing node doesn't participate in a cycle
                yield [self._nodes[index] for index in components]

    def sort(self) -> Iterator[AttributeNode]:
        """
        Sorts the graph in reverse topological order and raises a :py:exc:`DependencyCycleError` if dependency cycles are detected

        :return: A generator over :py:class:`AttributeNode` instance in reverse topological order according to their dependencies
        """
        for component in self.strongly_connected_components():
            if len(component) > 1:
                raise DependencyCycleError(f"Dependency cycle detected: {' -> '.join(node.name for node in component)}")
            yield component[0]


@marshmallow_dataclass.dataclass
class _AttributeGraphSchema:
    """
    Dataclass schema for :py:class:`AttributeGraph` serialization

    :param nodes: The phonetic attribute nodes of the graph
    :param node_indices: A mapping of node names to node indices
    :param edges: A mapping of node indices to the indices of their dependents
    """

    Schema: ClassVar[Type[Schema]]

    nodes: List[AttributeNode]
    node_indices: Dict[str, int]
    edges: List[List[int]]


class AttributeGraphField(Field):
    """
    :py:mod:`marshmallow` field for serializing and deserializing an :py:class:`AttributeGraph`
    """

    _Schema = _AttributeGraphSchema.Schema()

    def _serialize(self, value: AttributeGraph, _attr: str, _obj: Any, **kwargs) -> Dict[str, Any]:
        return self._Schema.dump(
            _AttributeGraphSchema(  # type: ignore
                value._nodes,
                value._node_indices,
                value._edges,
            )
        )

    def _deserialize(self, value: Any, _attr: str | None, _data: Mapping[str, Any] | None, **kwargs) -> AttributeGraph:
        fields = typing.cast(_AttributeGraphSchema, self._Schema.load(value))
        # Manually initializes the graph from the deserialized fields
        graph = AttributeGraph.__new__(AttributeGraph)
        graph._nodes = fields.nodes
        graph._node_indices = fields.node_indices
        graph._edges = fields.edges
        return graph



from dataclasses import dataclass
from enum import Enum
import json
import os
import sys
from argparse import Action, ArgumentParser, ArgumentTypeError, FileType, Namespace
from collections.abc import Iterable, Iterator, Sequence
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar
from marshmallow import fields
import logging
import marshmallow_dataclass

import toml
import torch
from torch import Tensor
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
import zarr
from zarr import Blosc
from zarr.errors import ContainsArrayError
import numpy as np

from allophant import utils
from allophant.config import Config, FeatureSet
from allophant.datasets.speech_corpus import (
    LanguageData,
    LanguageInfo,
    MultilingualCorpus,
    MultilingualSplits,
    TranscribedUtterance,
)
from allophant.package_data import DEFAULT_CONFIG_PATH
from allophant.phoneme_segmentation import SEGMENTATION_LOGGER
from allophant.phonetic_features import PhoneticAttributeIndexer
from allophant.preprocessing import FeatureFunction
from allophant.utils import EnumAction, MarshmallowDataclassLoadMixin
from allophant.datasets import corpus_loading, mozilla_common_voice
from allophant.datasets.corpus_loading import PreprocessedSplitData
from allophant.datasets.mozilla_common_voice import CommonVoiceCorpus, ReleaseMeta
from allophant.datasets.phonemes import G2PEngineType, GraphemeToPhonemeEnsemble
from allophant.language_codes import LanguageCodeMap


COMMON_VOICE_LOGGER = logging.getLogger("allophant.datasets")
COMMON_VOICE_LOGGER.setLevel(logging.INFO)


def generate_phoneme_transcriptions(parser: ArgumentParser, arguments: Namespace) -> None:
    if arguments.engine is None:
        g2p_engine = None

        if arguments.feature_set or arguments.attribute_path:
            parser.error(
                "Attribute path (-a/--attribute-path) and/or feature sets (-f/--feature-set) are only valid when a grapheme to phoneme engine is specified"
            )
    else:
        if arguments.feature_set:
            phoneme_inventory = PhoneticAttributeIndexer(
                arguments.feature_set, arguments.attribute_path
            ).phonemes.tolist()
        elif arguments.attribute_path:
            with arguments.attribute_path as file:
                phoneme_inventory = [segment.strip() for segment in file]
        else:
            parser.error(
                "At least one of -f/--feature-set and -a/--attribute-path has to be specified if grapheme to phoneme engines are given"
            )

        existing_engines = set()
        engines = []
        for engine_type in arguments.engine:
            # Sanity check since arguments might be passed from another source than __main__
            if engine_type in existing_engines:
                parser.error(f"G2P engine {engine_type!r} specified more than once")

            engines.append(engine_type.model())
            existing_engines.add(engine_type)

        g2p_engine = GraphemeToPhonemeEnsemble(engines, phoneme_inventory, arguments.batch_size)
        if arguments.log is not None:
            SEGMENTATION_LOGGER.add_file_handler(arguments.log)

    has_language_codes = arguments.language_codes is not None
    has_language_limits = isinstance(arguments.training_limits, dict)
    if has_language_codes or has_language_limits:
        language_map = LanguageCodeMap(
            ReleaseMeta.load(os.path.join(arguments.common_voice_path, CommonVoiceCorpus.META_FILE)).language_codes()
        )
        if has_language_codes:
            arguments.language_codes = [language_map[code] for code in arguments.language_codes]
        if has_language_limits:
            arguments.training_limits = {language_map[code]: limit for code, limit in arguments.training_limits.items()}

    mozilla_common_voice.load_common_voice(
        arguments.common_voice_path,
        g2p_engine,
        arguments.feature_set or arguments.attribute_path,
        arguments.include_single_upvote,
        arguments.batch_size,
        arguments.language_codes,
        arguments.map_to_allophoible,
        arguments.training_limits,
        arguments.progress,
        arguments.mapping_threshold,
    ).save(arguments.common_voice_path, arguments.output_directory)


def download_meta(_: ArgumentParser, arguments: Namespace) -> None:
    with arguments.output as output:
        json.dump(mozilla_common_voice.download_release_meta(arguments.version), output)


def parse_limits(limits: str) -> int | Dict[str, int]:
    try:
        return int(limits)
    except ValueError:
        return fields.Dict(fields.String(), fields.Integer()).deserialize(json.loads(limits))  # type: ignore


I = TypeVar("I", bound=LanguageInfo)
T = TypeVar("T", bound=TranscribedUtterance)


class FeaturePreprocessingDataset(Dataset, Generic[I, T]):
    def __init__(
        self, corpus: MultilingualCorpus[I, T], language_data: LanguageData[I, T], feature_function: FeatureFunction
    ) -> None:
        self._corpus = corpus
        self._samples = language_data.transcribed_samples
        self._language = language_data.info.code
        self._feature_function = feature_function

    def __getitem__(self, index: int) -> Tensor:
        # Get left channel only (or single channel for mono audio)
        return self._feature_function(
            self._corpus.audio_from_utterance(self._language, self._samples[index].utterance_id)[0][0]
        )

    def __len__(self) -> int:
        return len(self._samples)


def _create_audio_batch(batch: List[Tensor]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Workaround for when subarrays happen to have the same size,
    which would yield a regular array instead of an object array
    """
    batch_size = len(batch)
    object_array = np.empty(batch_size, dtype="object")
    lengths = np.empty(batch_size)
    for i, samples in enumerate(batch):
        lengths[i] = samples.shape[0]
        object_array[i] = samples.flatten().numpy()

    return object_array, lengths


_FEATURE_CHUNK_SIZE = 128


def preprocess_filters(_: ArgumentParser, arguments: Namespace) -> None:
    config = Config.load(arguments.config)
    corpus = corpus_loading.load_corpus(
        arguments.dataset_path,
        arguments.dataset_type,
        config.preprocessing.resample,
        only_primary_script=config.data.only_primary_script,
        progress_bar=True,
    )
    feature_function = FeatureFunction.from_config(config, corpus.audio_info().sample_rate)
    data_workers = utils.get_worker_count(arguments.data_workers)

    store = zarr.DirectoryStore(arguments.output_directory)
    data_root = zarr.group(store)
    # Store metadata for unraveling tensors later
    data_root.attrs["feature_size"] = feature_function.feature_size
    split_groups = data_root.require_groups(*corpus.SPLIT_NAMES)
    with tqdm(total=sum(map(len, corpus)), unit=" files") as progress:
        for corpus_split, split_group in zip(corpus, split_groups):
            for language_data in corpus_split:
                language_group = split_group.require_group(language_data.info.code)
                # Saves preprocessed features in flattened variable length arrays
                feature_data = language_group.zeros(
                    "features",
                    shape=len(language_data),
                    chunks=_FEATURE_CHUNK_SIZE,
                    compressor=Blosc(cname="lz4", shuffle=Blosc.BITSHUFFLE),
                    dtype="array:float32",
                )
                try:
                    length_data = language_group.zeros(
                        "lengths", shape=len(language_data), chunks=_FEATURE_CHUNK_SIZE, dtype=np.int64
                    )
                # Accept lengths already being generated by save-lengths
                except ContainsArrayError:
                    length_data = None

                offset = 0
                for features, lengths in DataLoader(
                    FeaturePreprocessingDataset(corpus_split, language_data, feature_function),
                    batch_size=_FEATURE_CHUNK_SIZE,
                    collate_fn=_create_audio_batch,
                    shuffle=False,
                    num_workers=data_workers,
                    persistent_workers=True,
                ):
                    length = len(features)
                    end_offset = offset + length
                    feature_data[offset:end_offset] = features
                    if length_data is not None:
                        length_data[offset:end_offset] = lengths
                    offset += length
                    progress.update(length)


def save_lengths(_: ArgumentParser, arguments: Namespace) -> None:
    config = Config.load(arguments.config)
    corpus = corpus_loading.load_corpus(
        arguments.dataset_path,
        arguments.dataset_type,
        config.preprocessing.resample,
        only_primary_script=config.data.only_primary_script,
        progress_bar=True,
    )
    feature_function = FeatureFunction.from_config(config, corpus.audio_info().sample_rate)

    store = zarr.DirectoryStore(arguments.output_directory)
    data_root = zarr.group(store)
    split_groups = data_root.create_groups(*corpus.SPLIT_NAMES)

    with tqdm(total=sum(map(len, corpus)), unit=" files") as progress:

        def _frame_counts(lengths: Iterable[int]) -> Iterator[int]:
            for length in lengths:
                yield feature_function.frame_count(length)
                progress.update(1)

        for split, split_group in zip(corpus, split_groups):
            for language_code, lengths in split.read_lengths():
                split_group.create_group(language_code).array(
                    "lengths", list(_frame_counts(lengths)), chunks=_FEATURE_CHUNK_SIZE, dtype=np.int64
                )


@dataclass
class UtteranceDurations:
    total: int
    average: float
    sample_rate: Optional[int] = None

    def __str__(self) -> str:
        if self.sample_rate is None:
            seconds = hours = 1
        else:
            seconds = self.sample_rate
            hours = self.sample_rate * 60 * 60
        return f"{self.total} frames at {self.sample_rate}: {self.total / hours:.2f}h (Average: {self.average / seconds:.4f}s)"

    @classmethod
    def compute_durations(cls, lengths: Tensor, sample_rate: int | None = None):
        return cls(
            int(lengths.sum().item()),
            lengths.mean(dtype=torch.float64).item(),
            sample_rate,
        )


@dataclass
class SplitStatistics:
    languages: List[str]
    utterance_counts: Dict[str, int]
    phoneme_statistics: Dict[str, Dict[str, int]]
    durations: Optional[Dict[str, UtteranceDurations]] = None


@marshmallow_dataclass.add_schema
@dataclass
class CorpusStatistics(MarshmallowDataclassLoadMixin):
    splits: Dict[str, SplitStatistics]


def _range_to_slice(input_range: range) -> slice:
    return slice(input_range.start, input_range.stop)


def _length_statistics(
    split: MultilingualCorpus, data: PreprocessedSplitData, sample_rate: int | None = None
) -> Dict[str, UtteranceDurations]:
    lengths = data.lengths
    return {
        language: UtteranceDurations.compute_durations(
            lengths[_range_to_slice(split.monolingual_index_range(language))], sample_rate
        )
        for language in split.languages
    }


def corpus_statistics(_: ArgumentParser, arguments: Namespace) -> None:
    corpus = corpus_loading.load_corpus(
        arguments.dataset_path,
        arguments.dataset_type,
        only_primary_script=arguments.only_primary_script,
        progress_bar=True,
    )

    statistics = CorpusStatistics({})
    for split_name in arguments.splits:
        split = getattr(corpus, split_name)
        statistics.splits[split_name] = SplitStatistics(split.languages.copy(), {}, {})

    if arguments.lengths is not None:
        data = corpus_loading.preprocessed_features_or_lengths(
            arguments.lengths, {split: getattr(corpus, split).languages for split in arguments.splits}
        )

        for split_name, split_data in data.items():
            statistics.splits[split_name].durations = _length_statistics(
                getattr(corpus, split_name), split_data, arguments.sample_rate
            )

    if arguments.json:
        print(statistics.dumps())
    else:
        print(statistics)


class EnumUniqueAppendAction(Action):
    """
    An action for handling sets of multiple Enum instances
    """

    def __init__(self, type: Type[Enum], **kwargs) -> None:
        self._enum_action = EnumAction(type, **kwargs)
        self._seen = set()
        super().__init__(**kwargs)

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        # Get current values
        current_values = getattr(namespace, self.dest) or []
        # Retrieve destination value from the enum action
        self._enum_action(parser, namespace, values, option_string)
        new_value = getattr(namespace, self.dest)
        # Ignore duplicates and append new elements
        if new_value not in self._seen:
            current_values.append(new_value)
            self._seen.add(new_value)
        setattr(namespace, self.dest, current_values)


def main(args: Sequence[str] | None = None) -> None:
    if args is None:
        args = sys.argv[1:]

    parser = ArgumentParser(description="Tools for preprocessing the Mozilla Common Voice corpus")
    subparsers = parser.add_subparsers(dest="mode")
    subparsers.required = True

    transcription_parser = subparsers.add_parser(
        "transcribe",
        help="Generates phonemic transcriptions for (a subset of) the Mozilla Common Voice corpus and extracts phoneme inventories",
        description="Generates phonemic transcriptions for (a subset of) the Mozilla Common Voice corpus and extracts phoneme inventories",
    )
    transcription_parser.add_argument("common_voice_path", help="Path to a version of the Mozilla Common Voice dataset")
    transcription_parser.add_argument(
        "-e",
        "--engine",
        action=EnumUniqueAppendAction,
        type=G2PEngineType,
        help="G2P engine(s) for extracting phonemic transcriptions for each utterance\nMay be none or multiple",
    )
    transcription_parser.add_argument(
        "-u",
        "--include-single-upvote",
        action="store_true",
        help="If given, unvalidated utterances with at least one more upvotes than downvotes will also be included in the training set",
    )
    transcription_parser.add_argument(
        "-b",
        "--batch-size",
        type=int,
        default=128,
        help="Batch size for tokenization and G2P engines (if provided) in maximum number of utterances per batch",
    )
    transcription_parser.add_argument(
        "-l",
        "--language-codes",
        type=lambda codes: codes.split(","),
        help="ISO639 language codes for each language to be transcribed",
    )
    transcription_parser.add_argument(
        "-t",
        "--training-limits",
        type=utils.argparse_type_wrapper(parse_limits),
        help="Limit for the number of utterances per language, may be a single number or a json dictionary from IOS639 language codes to language specific limits",
    )
    transcription_parser.add_argument(
        "-a",
        "--attribute-path",
        type=FileType("r", encoding="utf-8"),
        help="Path to a phoneme inventory with one phoneme per line or a supported phoneme feature file",
    )
    transcription_parser.add_argument(
        "-f",
        "--feature-set",
        action=EnumAction,
        type=FeatureSet,
        help="Type of feature set to load - loads Allophoible for the phoible feature set unless --attribute-path is given",
    )
    transcription_parser.add_argument(
        "-p", "--progress", action="store_true", help="Shows progress bars while utterances are being transcribed"
    )
    transcription_parser.add_argument(
        "-o",
        "--output-directory",
        help=(
            "A different directory from the corpus directory for storing transcriptions."
            "Note that moving the original corpus directory makes audio files inaccessible from this folder."
        ),
    )
    transcription_parser.add_argument(
        "-m",
        "--map-to-allophoible",
        action="store_true",
        help=(
            "Maps the inventories collected from the G2P transcriptions to language specific allophoible inventories, "
            "including in the transcriptions themselves"
        ),
    )
    transcription_parser.add_argument(
        "--log", help="Path to the log file for phonemes missing from the chosen feature set"
    )
    transcription_parser.add_argument(
        "--mapping-threshold",
        type=int,
        help="Distance threshold for allophoible mapping above which mapping should not take place",
    )

    download_meta_parser = subparsers.add_parser(
        "download-meta",
        help="Downloads metadata for a given version of the Mozilla Common Voice corpus",
        description="Downloads metadata for a given version of the Mozilla Common Voice corpus",
    )
    download_meta_parser.add_argument(
        "output",
        type=FileType("w", encoding="utf-8"),
        default=sys.stdout,
        help="Output file which the metadata will be saved to",
    )
    download_meta_parser.add_argument(
        "-v", "--version", default="10.0-2022-07-04", help="Version of the Common Voice corpus"
    )

    data_config_parser = ArgumentParser(add_help=False)
    data_config_parser.add_argument(
        "-t", "--dataset-type", choices={"common-voice"}, default="common-voice", help="Type of the Dataset"
    )
    data_config_parser.add_argument(
        "-c",
        "--config",
        type=utils.argparse_type_wrapper(toml.load),
        default=toml.load(DEFAULT_CONFIG_PATH),
        help=(
            "Path to a configuration file in toml format, "
            "usually modified from a default config generated using generate-config"
        ),
    )
    data_config_parser.add_argument(
        "-j",
        "--config-json-data",
        default=None,
        help="Parses configuration files passed to `-c/--config` using JSON instead of TOML",
    )

    data_processing_parser = ArgumentParser(add_help=False)
    data_processing_parser.add_argument(
        "dataset_path", help="Path to a corpus containing phonetically transcribed utterances"
    )
    data_processing_parser.add_argument(
        "output_directory", help="Path to a directory which the processed data will be stored in"
    )

    subparsers.add_parser(
        "save-lengths",
        parents=[data_config_parser, data_processing_parser],
        help="Generates and saves frame lengths for the given dataset based on the feature function in the given configuration",
        description="Generates and saves frame lengths for the given dataset based on the feature function in the given configuration",
    )

    preprocessing_parser = subparsers.add_parser(
        "preprocess",
        parents=[data_config_parser, data_processing_parser],
        help="Preprocesses acoustic features for the given dataset based on the given configuration",
        description="Preprocesses acoustic features for the given dataset based on the given configuration",
    )
    preprocessing_parser.add_argument(
        "-w",
        "--data-workers",
        type=int,
        help=(
            "Number of workers - 0 disables workers and runs data processing synchronously on the main thread, "
            "by default the number of workers is auto-detected based on CPU thread count"
        ),
    )

    statistics_parser = subparsers.add_parser("stats")
    statistics_parser.add_argument(
        "dataset_path", help="Path to a corpus containing phonetically transcribed utterances"
    )
    statistics_parser.add_argument(
        "-t", "--dataset-type", choices={"common-voice"}, default="common-voice", help="Type of the Dataset"
    )
    statistics_parser.add_argument(
        "-l", "--lengths", help="Path to previously saved lengths for utterances in the corpus"
    )
    statistics_parser.add_argument(
        "-p",
        "--only-primary-script",
        action="store_true",
        help="Keeps only transcriptions that consist of a single script that corresponds to the primary script of each language",
    )
    statistics_parser.add_argument(
        "-s",
        "--splits",
        type=lambda string: string.split(","),
        default=MultilingualSplits.SPLIT_NAMES,
        help="Splits to calculate statistics for separated by commas. E.g.: train,dev,test",
    )
    statistics_parser.add_argument("-j", "--json", action="store_true", help="Outputs statistics in JSON format")
    statistics_parser.add_argument(
        "--sample-rate",
        type=int,
        help="Interprets lengths as frame counts at the given sample rate, if provided, and displays durations also in hours",
    )

    arguments = parser.parse_args(args)

    # Handles cases in which the configuration is provided in json format
    if hasattr(arguments, "config_json_data") and arguments.config_json_data is not None:
        arguments.config = json.loads(arguments.config_json_data)

    match arguments.mode:
        case "transcribe":
            generate_phoneme_transcriptions(parser, arguments)
        case "download-meta":
            download_meta(parser, arguments)
        case "save-lengths":
            save_lengths(parser, arguments)
        case "preprocess":
            preprocess_filters(parser, arguments)
        case "stats":
            corpus_statistics(parser, arguments)
        case mode:
            raise ValueError(f"Unsupported action: {mode}")


if __name__ == "__main__":
    main(sys.argv[1:])



from typing import TypeVar, Type, Tuple, Callable, Optional, Sequence, Union
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum

import torch
from torch import Tensor
from torch import nn
from torch.nn import functional

from allophant import utils
from allophant.dataset_processing import Batch
from allophant.config import (
    Glu1dConfig,
    DropoutConfig,
    LayerNormConfig,
    MaxPoolingConfig,
    FrontendConfig,
    DirectFrontendConfig,
    LinearFrontendConfig,
    SequentialFrontendConfig,
)
from allophant.network.padding import VariableLengthReflectPad
from allophant.network import padding


T = TypeVar("T", bound=Union[float, Tensor])


class ShapeMode(Enum):
    LENGTH_FIRST = 0
    BATCH_FIRST = 1


@dataclass
class MaskInfo:
    pattern: Sequence[int] = (-1, -1)
    shape_mode: ShapeMode = ShapeMode.BATCH_FIRST

    def __post_init__(self):
        self._mask_pattern = tuple(slice(None) if dim == -1 else None for dim in self.pattern)

    def to_shape(self, mask: Tensor) -> Tensor:
        if self.shape_mode == ShapeMode.LENGTH_FIRST:
            mask = mask.T
        return mask[self._mask_pattern]


class LengthWrapper(nn.Module):
    def __init__(
        self,
        module: nn.Module,
        length_function: Optional[Callable[[Tensor], Tensor]] = None,
        mask_info: Optional[MaskInfo] = None,
        pass_length_info: bool = False,
    ):
        super().__init__()
        self._length_function = length_function
        self._mask_input = mask_info
        self._pass_length_info = pass_length_info
        self.module = module

    def forward(self, inputs: Tensor, lengths: Tensor) -> Tuple[Tensor, Tensor]:
        if self._length_function is None:
            if self._pass_length_info:
                return self.module(inputs, lengths), lengths
            return self.module(inputs), lengths

        if self._mask_input is not None:
            inputs *= self._mask_input.to_shape(utils.mask_sequence(lengths))
        out_lengths = self._length_function(lengths)
        if self._pass_length_info:
            return self.module(inputs, lengths), out_lengths
        return self.module(inputs), out_lengths

    def lengths(self, lengths: Tensor) -> Tensor:
        if self._length_function is None:
            return lengths
        return self._length_function(lengths)


class LengthSequential(nn.Module):
    def __init__(self, *args: LengthWrapper):
        super().__init__()
        self.layers = nn.ModuleList(args)

    def forward(self, inputs: Tensor, lengths: Tensor) -> Tuple[Tensor, Tensor]:
        for layer in self.layers:
            inputs, lengths = layer(inputs, lengths)
        return inputs, lengths

    def lengths(self, lengths: Tensor) -> Tensor:
        for layer in self.layers:
            lengths = layer.lengths(lengths)  # type: ignore
        return lengths


class Glu1d(nn.Module):
    def __init__(
        self,
        input_dimensions: int,
        output_dimensions: int,
        kernel_size: int,
        stride: int = 1,
        reflect_pad: bool = True,
    ):
        super().__init__()
        self._padding = padding.get_padding(kernel_size, stride)
        self._reflect_padding = VariableLengthReflectPad(self._padding) if reflect_pad else None
        self._kernel_size = kernel_size
        self._stride = stride
        self._weights = nn.Conv1d(
            input_dimensions,
            output_dimensions * 2,
            kernel_size=kernel_size,
            stride=stride,
        )

    @property
    def padding(self) -> Tuple[int, int]:
        return self._padding

    @property
    def kernel_size(self) -> int:
        return self._kernel_size

    @property
    def stride(self) -> int:
        return self._stride

    def forward(self, inputs: Tensor, lengths: Tensor) -> Tensor:
        if self._reflect_padding is None:
            inputs = functional.pad(inputs, self._padding)
        else:
            inputs = self._reflect_padding(inputs, lengths)
        return functional.glu(self._weights(inputs), 1)


class Frontend(nn.Module, metaclass=ABCMeta):
    _output_dimensions: int

    @abstractmethod
    def forward(self, batch: Batch) -> Batch:
        pass

    @property
    def output_dimensions(self) -> int:
        return self._output_dimensions

    def lengths(self, input_lengths: Tensor) -> Tensor:
        return input_lengths


class DirectFrontend(Frontend):
    def __init__(self, config: DirectFrontendConfig, feature_size: int):
        super().__init__()
        self._output_dimensions = feature_size
        self._dropout = nn.Dropout(config.input_dropout) if config.input_dropout > 0 else None

    def forward(self, batch: Batch) -> Batch:
        output = batch.audio_features
        if self._dropout is not None:
            output = self._dropout(output)
        return Batch(output, batch.lengths, batch.language_ids)


class LinearFrontend(Frontend):
    def __init__(self, config: LinearFrontendConfig, feature_size: int, elementwise_affine: bool = False):
        super().__init__()
        self._output_dimensions = config.neurons
        linear = nn.Linear(feature_size, config.neurons)
        if config.input_dropout > 0:
            self._layer = nn.Sequential(
                nn.Dropout(config.input_dropout),
                nn.LayerNorm(feature_size, elementwise_affine=elementwise_affine),
                linear,
                nn.LeakyReLU(),
            )
        else:
            self._layer = nn.Sequential(
                nn.LayerNorm(feature_size, elementwise_affine=elementwise_affine),
                linear,
                nn.LeakyReLU(),
            )

    def forward(self, batch: Batch) -> Batch:
        return Batch(
            self._layer(batch.audio_features.transpose(1, 2)).transpose(1, 2), batch.lengths, batch.language_ids
        )


def conv_length(
    kernel_size: int, stride: int = 1, use_padding: bool = True, stft_type: bool = False
) -> Callable[[Tensor], Tensor]:
    if use_padding:
        layer_padding = sum(padding.get_padding(kernel_size, stride, stft_type))
    else:
        layer_padding = 0

    def padded_length(lengths: Tensor) -> Tensor:
        return torch.div(((lengths + layer_padding) - kernel_size), stride, rounding_mode="floor") + 1

    return padded_length


class Transpose(nn.Module):
    def __init__(self, dimension_a: int, dimension_b: int):
        super().__init__()
        self._dimension_a = dimension_a
        self._dimension_b = dimension_b

    def forward(self, inputs: Tensor) -> Tensor:
        return inputs.transpose(self._dimension_a, self._dimension_b)


SequentialFrontendCls = TypeVar("SequentialFrontendCls", bound="SequentialFrontend")


class SequentialFrontend(Frontend):
    def __init__(self, layers: LengthSequential, output_dimensions: int, upscale_factor: float = 1):
        super().__init__()
        self._layers = layers
        self._output_dimensions = output_dimensions
        self._upscale_factor = upscale_factor

    @classmethod
    def from_config(
        cls: Type[SequentialFrontendCls], config: SequentialFrontendConfig, feature_size: int
    ) -> SequentialFrontendCls:
        layers = []
        previous_output_size = feature_size
        upscale_factor = 1
        for layer in config.layers:
            if isinstance(layer, DropoutConfig):
                layers.append(LengthWrapper(nn.Dropout(layer.rate)))
            elif isinstance(layer, Glu1dConfig):
                output_size = layer.out_channels
                module = Glu1d(previous_output_size, output_size, layer.kernel, layer.stride)
                layers.append(
                    LengthWrapper(
                        module,
                        conv_length(module.kernel_size, module.stride),
                        MaskInfo((-1, 1, -1)),
                        pass_length_info=True,
                    )
                )
                previous_output_size = output_size
                upscale_factor *= module.stride
            elif isinstance(layer, LayerNormConfig):
                layers.append(
                    LengthWrapper(
                        nn.Sequential(
                            Transpose(-1, -2),
                            nn.LayerNorm(previous_output_size, elementwise_affine=layer.affine),
                            Transpose(-2, -1),
                        )
                    )
                )
            elif isinstance(layer, MaxPoolingConfig):
                layers.append(LengthWrapper(nn.MaxPool1d(layer.size), conv_length(layer.size), MaskInfo((-1, 1, -1))))
                upscale_factor *= layer.size
            else:
                raise ValueError(f"Unsupported layer config of type: {layer.__class__.__name__}")

        return cls(LengthSequential(*layers), previous_output_size, upscale_factor)

    @property
    def upscale_factor(self) -> float:
        return self._upscale_factor

    def forward(self, batch: Batch) -> Batch:
        features, lengths = self._layers(batch.audio_features, batch.lengths)
        return Batch(features, lengths, batch.language_ids)

    def downsampled_lengths(self, lengths: Tensor) -> Tensor:
        return self._layers.lengths(lengths)


def frontend_from_config(
    frontend_config: FrontendConfig, feature_size: int, elementwise_affine: bool = False
) -> Frontend:
    if isinstance(frontend_config, DirectFrontendConfig):
        return DirectFrontend(frontend_config, feature_size)
    if isinstance(frontend_config, LinearFrontendConfig):
        return LinearFrontend(frontend_config, feature_size, elementwise_affine)

    raise ValueError(f"Unsupported frontend config type {frontend_config.__class__.__name__}")



from abc import ABCMeta
from collections.abc import Iterable, Iterator, Set
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Generic, List, Tuple, TypeVar

from torch import LongTensor, Tensor
import torch
from torch.utils.data import Dataset
import numpy as np

from allophant.config import Config
from allophant.datasets.speech_corpus import (
    IndexedEntry,
    LanguageInfo,
    MultilingualCorpus,
    PhoneticallySegmentedUtterance,
    PhoneticallyTranscribedUtterance,
)
from allophant.phonetic_features import PhonemeIndexer, PhoneticAttributeIndexer, PhoneticIndexerState
from allophant.preprocessing import FeatureFunction


@dataclass
class SamplesProcessor:
    feature_function: FeatureFunction
    attribute_indexer: PhoneticAttributeIndexer

    @property
    def feature_size(self) -> int:
        return self.feature_function.feature_size

    def indexer_state(self) -> PhoneticIndexerState:
        return self.attribute_indexer.state()

    @classmethod
    def from_config(
        cls,
        config: Config,
        sampling_rate: int,
        attribute_indexer: PhoneticAttributeIndexer,
    ):
        return cls(
            FeatureFunction.from_config(config, sampling_rate),
            attribute_indexer,
        )


@dataclass
class Batch:
    audio_features: Tensor
    lengths: Tensor
    language_ids: Tensor

    def pin_memory(self):
        self.audio_features = self.audio_features.pin_memory()
        self.lengths = self.lengths.pin_memory()
        self.language_ids = self.language_ids.pin_memory()
        return self

    def _inputs_to(
        self, device: torch.device | str | None, non_blocking: bool = False, copy: bool = False
    ) -> Tuple[Tensor, Tensor, Tensor]:
        return (
            self.audio_features.to(device, non_blocking=non_blocking, copy=copy),
            self.lengths.to(device, non_blocking=non_blocking, copy=copy),
            self.language_ids.to(device, non_blocking=non_blocking, copy=copy),
        )

    def to(self, device: torch.device | str | None, non_blocking: bool = False, copy: bool = False):
        return self.__class__(*self._inputs_to(device, non_blocking, copy))

    def cuda(self, non_blocking: bool = False, copy: bool = False):
        return self.to("cuda", non_blocking, copy)

    def size(self) -> int:
        return len(self)

    def __len__(self) -> int:
        return self.lengths.numel()

    def __repr__(self) -> str:
        return "{}(Features: ({}; {}))".format(
            self.__class__.__name__, self.audio_features.shape, self.audio_features.dtype
        )


RawLabeledBatchCls = TypeVar("RawLabeledBatchCls", bound="RawLabeledBatch")


@dataclass
class RawLabeledBatch(Batch):
    raw_labels: List[List[List[str]]]
    utterance_ids: List[str]

    def to(self, device: torch.device | str | None, non_blocking: bool = False, copy: bool = False):
        return self.__class__(
            *self._inputs_to(device, non_blocking, copy),
            self.raw_labels,
            self.utterance_ids,
        )

    def split_by_language(self: RawLabeledBatchCls) -> Iterator[Tuple[int, RawLabeledBatchCls]]:
        split_ids, split_indices = self.language_ids.unique_consecutive(return_counts=True)
        # Turn sizes into end indices for tensor split
        split_indices.cumsum_(0)

        offset = 0
        for split_id, split_index, features, lengths, language_ids in zip(
            split_ids,
            split_indices,
            self.audio_features.tensor_split(split_indices),
            self.lengths.tensor_split(split_indices),
            self.language_ids.tensor_split(split_indices),
        ):
            yield (
                split_id,
                self.__class__(
                    features[..., : lengths.max()],
                    lengths,
                    language_ids,
                    [labels[offset:split_index] for labels in self.raw_labels],
                    self.utterance_ids[offset:split_index],
                ),
            )
            offset = split_index


LabeledBatchCls = TypeVar("LabeledBatchCls", bound="LabeledBatch")


@dataclass
class LabeledBatch(Batch):
    attribute_indices: List[Dict[str, Tensor]]
    label_lengths: List[Tensor]
    label_length_indices: Dict[str, int]

    def pin_memory(self):
        super().pin_memory()
        for lengths in self.label_lengths:
            lengths.pin_memory()

        for engine in self.attribute_indices:
            for indices in engine.values():
                indices.pin_memory()

        return self

    def to(self, device: torch.device | str | None, non_blocking: bool = False, copy: bool = False):
        return self.__class__(
            *self._inputs_to(device, non_blocking, copy),
            [
                {
                    name: labels.to(device, non_blocking=non_blocking, copy=copy)
                    for name, labels in engine.items()
                    if isinstance(labels, Tensor)
                }
                for engine in self.attribute_indices
            ],
            [tensor.to(device, non_blocking=non_blocking, copy=copy) for tensor in self.label_lengths],
            label_length_indices=self.label_length_indices,
        )


class BatchType(Enum):
    UNLABELED = 0
    RAW = 1
    INDEXED = 2


I = TypeVar("I", bound=LanguageInfo)
T = TypeVar("T", bound=PhoneticallyTranscribedUtterance | PhoneticallySegmentedUtterance)
B = TypeVar("B", bound=Batch)


class PhonemeDataset(Dataset, Generic[I, T, B], metaclass=ABCMeta):
    _corpus: MultilingualCorpus[I, T]
    _processor: SamplesProcessor
    _index_start_offset: int
    _features: List[np.ndarray] | None
    _indexer: PhonemeIndexer
    _batch_type: BatchType
    _inventories: Dict[int, Set[str]] | None = None

    def phoneme_count(self) -> int:
        return len(self._processor.attribute_indexer)

    @property
    def corpus(self) -> MultilingualCorpus[I, T]:
        return self._corpus

    @property
    def unrestricted_inventory(self) -> bool:
        return True

    @property
    def indexer(self) -> PhonemeIndexer:
        return self._indexer

    @property
    def batch_type(self) -> BatchType:
        return self._batch_type

    def __len__(self) -> int:
        return len(self._corpus)

    def _filter_with_inventory(self, language_id: int, segmented_entries: Iterable[List[str]]) -> Iterable[List[str]]:
        inventory = self._inventories
        if inventory is None:
            return segmented_entries

        inventory = inventory[language_id]
        return (
            [phoneme for phoneme in segmented_sentence if phoneme in inventory]
            for segmented_sentence in segmented_entries
        )

    def _create_processed_labeled_batch(
        self, language_id: int, segmented_entries: Iterable[List[str]], features: Tensor, feature_size: Tensor
    ) -> LabeledBatch:
        engine_attribute_indices = []
        for segmented_sentence in segmented_entries:
            # Add start offset e.g. for CTC blanks
            engine_attribute_indices.append(
                self._indexer.get_named(self._indexer.phoneme_indices(segmented_sentence), self._index_start_offset)
            )

        return LabeledBatch(
            features,
            feature_size,
            torch.tensor(language_id),
            engine_attribute_indices,
            [LongTensor(list(map(len, attribute_indices.values()))) for attribute_indices in engine_attribute_indices],
            {name: index for index, name in enumerate(engine_attribute_indices[0])} if engine_attribute_indices else {},
        )

    def _generate_sample(
        self,
        index: int,
        transcribed_item: IndexedEntry[T],
        segmented_entries: Iterable[List[str]],
    ) -> B:
        language_id = transcribed_item.language_id

        if self._features is None:
            features = self._processor.feature_function(self._corpus.audio_for(transcribed_item)[0])
        else:
            # Uses torch.tensor instead of torch.from_numpy since creating tensors from non-writable numpy arrays can result in undefined behavior
            features = torch.tensor(self._features[index])

        feature_size = torch.tensor(features.shape[0])

        match self._batch_type:
            case BatchType.UNLABELED:
                return Batch(features, feature_size, torch.tensor(language_id))  # type: ignore
            case BatchType.INDEXED:
                return self._create_processed_labeled_batch(  # type: ignore
                    language_id, self._filter_with_inventory(language_id, segmented_entries), features, feature_size
                )
            case BatchType.RAW:
                return RawLabeledBatch(  # type: ignore
                    features,
                    feature_size,
                    torch.tensor(language_id),
                    [
                        [entry]
                        for entry in self._filter_with_inventory(
                            language_id,
                            segmented_entries,
                        )
                    ],
                    [transcribed_item.entry.utterance_id],
                )


P = TypeVar("P", bound=PhoneticallyTranscribedUtterance)


class TranscribedDataset(Generic[I, P, B], PhonemeDataset[I, P, B]):
    def __init__(
        self,
        batch_type: BatchType,
        corpus: MultilingualCorpus[I, P],
        processor: SamplesProcessor,
        index_start_offset: int = 0,
        features: List[np.ndarray] | None = None,
        inventories: Dict[int, Set[str]] | None = None,
        unrestricted_inventory: bool = False,
    ) -> None:
        self._batch_type = batch_type
        self._corpus = corpus
        self._processor: SamplesProcessor = processor
        self._index_start_offset = index_start_offset
        self._features = features
        self._inventories = inventories
        self._unrestricted_inventory = unrestricted_inventory
        self._indexer = (
            self._processor.attribute_indexer.full_subset_attributes
            if unrestricted_inventory
            else self._processor.attribute_indexer
        )

    @property
    def unrestricted_inventory(self) -> bool:
        return self._unrestricted_inventory

    def __getitem__(self, index: int) -> B:
        transcribed_item = self._corpus[index]
        # Flattens phonemes from each subsequence of the transcription
        transcribed = transcribed_item.entry.phonemes.flattened_transcriptions()

        return self._generate_sample(
            index,
            transcribed_item,
            transcribed,
        )


S = TypeVar("S", bound=PhoneticallySegmentedUtterance)


class PhoneticallySegmentedDataset(Generic[I, S, B], PhonemeDataset[I, S, B]):
    def __init__(
        self,
        batch_type: BatchType,
        corpus: MultilingualCorpus[I, S],
        processor: SamplesProcessor,
        index_start_offset: int = 0,
        features: List[np.ndarray] | None = None,
        inventories: Dict[int, Set[str]] | None = None,
    ) -> None:
        self._batch_type = batch_type
        self._corpus = corpus
        self._processor = processor
        self._index_start_offset = index_start_offset
        self._features = features
        self._inventories = inventories
        # Always use unrestricted target inventory for already segment datasets
        self._indexer = self._processor.attribute_indexer.full_subset_attributes

    def __getitem__(self, index: int) -> B:
        transcribed_item = self._corpus[index]
        return self._generate_sample(
            index,
            transcribed_item,
            (transcribed_item.entry.phonemes,),
        )



from abc import ABCMeta
import typing
from typing import (
    ClassVar,
    Type,
    TypeVar,
    Iterable,
    Dict,
    Protocol,
    Optional,
    Any,
    Sequence,
    Mapping,
    Generic,
    List,
    Union,
    Tuple,
    runtime_checkable,
)
import dataclasses
from dataclasses import dataclass
from enum import Enum
import re
from re import Pattern

import toml
from marshmallow import Schema
from marshmallow.exceptions import ValidationError
from marshmallow.fields import Nested, Field, String
from marshmallow.validate import OneOf, Range
from marshmallow import fields
from marshmallow_oneofschema import OneOfSchema
from marshmallow_enum import EnumField
import marshmallow_dataclass
from torch import optim
from torch.optim import Optimizer as TorchOptimizer
from torch import nn

from allophant.loss_functions import LossWrapper, CTCWrapper, SequenceCrossEntropyWrapper
from allophant import utils


F = TypeVar("F")


@runtime_checkable
class KeyedClass(Protocol):
    """Protocol for classes that have a TYPE class variable indicating its key in configs"""

    TYPE: ClassVar[str]


class KeyedOneOfSchema(OneOfSchema):
    """
    A Schema for (de)serializing a union of types tagged with a key field
    """

    types: ClassVar[Tuple[Type[KeyedClass], ...]]

    def get_obj_type(self, obj: object) -> str:
        if isinstance(obj, self.types):
            return obj.TYPE
        raise ValidationError(f"Unknown object type: {obj!r}")


def _choices(keyed_schemas: Iterable[Type[KeyedClass]]) -> Dict[str, Type[Schema]]:
    return {schema.TYPE: marshmallow_dataclass.class_schema(schema) for schema in keyed_schemas}


def _nested_field(schema: Type[Schema], field_args: Dict[str, Any] | None = None, **kwargs) -> Any:
    return utils.schema_field(Nested(schema, **(field_args if field_args is not None else {})), **kwargs)


def _nested_list_field(schema: Type[Schema], field_args: Dict[str, Any] | None = None, **kwargs) -> Any:
    return utils.schema_field(fields.List(Nested(schema, **(field_args if field_args is not None else {}))), **kwargs)


@dataclass
class WarmupInfo:
    """
    Configures learning rate warmup for transformers as derived
    from https://nlp.seas.harvard.edu/2018/04/03/attention.html#optimizer

    :param model_size: The input and output size of transformer blocks used in
        warmup calculations (`d_model`)
    """

    model_size: int


@marshmallow_dataclass.add_schema
@dataclass
class _WarmupState:
    """
    Contains the current step and learning rate for learning rate warmup

    :param step: current warmup step
    :param rate: current learning rate
    """

    Schema: ClassVar[Type[Schema]]

    step: int = 0
    rate: float = 0


class WarmupScheduler:
    _warmup_state: _WarmupState

    def __init__(
        self,
        optimizer: TorchOptimizer,
        warmup_info: WarmupInfo,
        warmup_steps: int,
        constant_steps: int = 0,
        factor: int = 2,
    ) -> None:
        self._optimizer = optimizer

        self._warmup_steps = warmup_steps
        self._constant_steps = constant_steps
        self._steps_until_decay = warmup_steps + constant_steps
        self._factor = factor
        self._model_size = warmup_info.model_size

        self._warmup_state = _WarmupState(1, 0)
        self._warmup_state.rate = self._rate(1)
        # Set initial learning rate
        self._set_lr(self._warmup_state.rate)

    def _set_lr(self, rate: float) -> None:
        for parameter in self._optimizer.param_groups:
            parameter["lr"] = rate

    @property
    def last_lr(self) -> float:
        return self._warmup_state.rate

    def _rate(self, step=None) -> float:
        """
        Computes the learning rate for the given warmup step
        as derived from https://nlp.seas.harvard.edu/2018/04/03/attention.html#optimizer

        :param step: Current step or `None` for the initial step

        :return: The new learning rate for the given step
        """
        # Based on implementation: https://nlp.seas.harvard.edu/2018/04/03/attention.html#optimizer
        if step is None:
            step = self._warmup_state.step

        # Warmup
        if step < self._warmup_steps:
            return self._factor * (self._model_size ** (-0.5) * (step * self._warmup_steps ** (-1.5)))
        # Keep constant at the maximum learning rate
        if step < self._steps_until_decay:
            return self._factor * (self._model_size ** (-0.5) * (self._warmup_steps ** (-0.5)))
        # Decay
        else:
            return self._factor * (self._model_size ** (-0.5) * ((step - self._constant_steps) ** (-0.5)))

    def step(self) -> None:
        # Based on implementation: https://nlp.seas.harvard.edu/2018/04/03/attention.html#optimizer
        self._warmup_state.step += 1
        rate = self._rate()
        self._set_lr(rate)
        self._warmup_state.rate = rate

    def state_dict(self) -> Dict[str, Any]:
        return {"warmup_state": _WarmupState.Schema().dump(self._warmup_state)}

    def load_state_dict(self, state_dict: Dict[str, Any]):
        self._warmup_state = _WarmupState.Schema().load(state_dict["warmup_state"])  # type: ignore


@dataclass
class WarmupConfig(KeyedClass):
    """
    Contains learning rate warmup parameters for transformers as derived
    from https://nlp.seas.harvard.edu/2018/04/03/attention.html#optimizer
    but with the addition of keeping the learning rate constant for a while until decaying

    :param warmup_steps: Number of batch updates until warmup completes
    :param constant_steps: The number of steps to keep the learning rate
        constant until decay. The schedule is equivalent to the standard
        transformer schedule if this is kept at 0
    :param factor: Scaling factor for warmup updates
    """

    TYPE: ClassVar[str] = "warmup"

    warmup_steps: int
    constant_steps: int = 0
    factor: int = 2

    def make_scheduler(self, optimizer: TorchOptimizer, warmup_info: WarmupInfo) -> WarmupScheduler:
        return WarmupScheduler(optimizer, warmup_info, self.warmup_steps, self.constant_steps, self.factor)


class LrSchedulerSchema(KeyedOneOfSchema):
    """Schema which selects between learning rate schedules"""

    type_field = "type"
    types = (WarmupConfig,)
    type_schemas = _choices(types)


# Lr schedulers that are available
LrSchedulerConfig = WarmupConfig


@dataclass
class OptimizerWrapper:
    """
    Wrapper around a `torch.Optimizer` with optional learning rate decay
    """

    def __init__(self, optimizer: TorchOptimizer, warmup_info: WarmupInfo):
        """
        Initializes a new optimizer wrapper with an optimizer and optional learning rate decay

        :param optimizer: A `torch.Optimizer`
        :param warmup_info: Information about the model required for configuring Warmup schedulers
        """
        self._optimizer = optimizer
        self._warmup_info = warmup_info

    def add_schedulers(self, scheduler_config: LrSchedulerConfig | None) -> None:
        """
        Adds a Learning Rate schedulers generated from the provided config

        :param schedule_configs: Configuration of a learning rate scheduler
        """
        self._lr_scheduler = (
            None if scheduler_config is None else scheduler_config.make_scheduler(self._optimizer, self._warmup_info)
        )

    @property
    def optimizer(self) -> TorchOptimizer:
        return self._optimizer

    def step(self) -> None:
        """
        Performs an optimization step and steps the learning rate scheduler if one exists
        """
        self.optimizer.step()
        if self._lr_scheduler is not None:
            self._lr_scheduler.step()

    @property
    def param_groups(self) -> List[Dict[Any, Any]]:
        return self._optimizer.param_groups

    def current_learning_rate(self) -> float:
        return self.param_groups[0]["lr"]

    def state_dict(self) -> Dict[str, Any]:
        """
        Returns the state of the OptimizerWrapper as a dict

        :return: The warmup, optimizer and scheduler states depending on configuration
        """
        return {
            "lr_scheduler": None if self._lr_scheduler is None else self._lr_scheduler.state_dict(),
            "optimizer": self._optimizer.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Any]):
        """
        Loads the optimizer wrapper state

        :param state_dict: The states of warmup, the optimizer and the schedulers depending on configuration.
            Should be a `dict` returned from a call to `state_dict`
        """
        self._optimizer.load_state_dict(state_dict["optimizer"])
        if self._lr_scheduler is not None:
            self._lr_scheduler.load_state_dict(state_dict["lr_scheduler"])


@dataclass
class Optimizer(metaclass=ABCMeta):
    """Superclass for all optimizer configurations which support learning rates and weight decay"""

    learning_rate: float
    l2_regularization: float = 0

    def get_optimizer(self, _params: Iterable[nn.parameter.Parameter], _warmup_info: WarmupInfo) -> OptimizerWrapper:
        """
        Retrieves an `OptimizerWrapper` based on the configuration options

        :param params: Parameters to be optimized
        :param warmup_info: Configuration and parameter information for learning rate warmup

        :return: An `OptimizerWrapper` optimizing `params`, optionall with learning rate warmup
        """
        raise NotImplementedError("No torch optimizer defined for this class")


@dataclass
class SGD(Optimizer, KeyedClass):
    """
    Configuration for the stochastic gradient descent (SGD) optimizer
    supporting weight decay, momentum and learning rate decay
    """

    TYPE: ClassVar[str] = "sgd"

    momentum: float = 0

    def get_optimizer(self, params: Iterable[nn.parameter.Parameter], warmup_info: WarmupInfo) -> OptimizerWrapper:
        sgd = optim.SGD(params, self.learning_rate, self.momentum, weight_decay=self.l2_regularization)
        return OptimizerWrapper(sgd, warmup_info)


@dataclass
class Adam(Optimizer, KeyedClass):
    """
    Configuration for the ADAM optimizer supporting weight and learning rate decay
    """

    TYPE: ClassVar[str] = "adam"

    learning_rate: float = 0.01
    beta_1: float = 0.9
    beta_2: float = 0.98

    def get_optimizer(self, params: Iterable[nn.parameter.Parameter], warmup_info: WarmupInfo) -> OptimizerWrapper:
        adam = optim.Adam(
            params, self.learning_rate, betas=(self.beta_1, self.beta_2), weight_decay=self.l2_regularization
        )
        return OptimizerWrapper(
            # Same betas as in post-lm and attention is all you need
            adam,
            warmup_info,
        )


class OptimizerSchema(KeyedOneOfSchema):
    """Schema which selects between optimizers"""

    type_field = "algorithm"
    types = (SGD, Adam)
    type_schemas = _choices(types)


@dataclass
class Window:
    """
    Window configuration for acoustic features operating on frames such as
    STFT-based features like Mel Filterbanks or MFCCs

    :param frame_duration: Duration of a frame in milliseconds
    :param frame_stride: Stride of the filter in milliseconds - also known as "hop"
    """

    frame_duration: int
    frame_stride: int

    @staticmethod
    def _ms_to_samples(milliseconds: int, sample_rate: int) -> int:
        """
        Converts milliseconds to the number of samples at the given sample rate

        :param milliseconds: Time in milliseconds to convert to samples
        :param sample_rate: Sample rate used in the conversion

        :return: Number of samples for the given duration at the given sample rate
        """
        return int((milliseconds / 1000) * sample_rate)

    def frame_duration_samples(self, sample_rate: int) -> int:
        """
        Calculates the frame duration in samples at the given sample rate

        :param sample_rate: Sample rate used for the conversion

        :return: The frame duration in samples at the given sample rate
        """
        return self._ms_to_samples(self.frame_duration, sample_rate)

    def frame_stride_samples(self, sample_rate: int) -> int:
        """
        Calculates the frame stride in samples at the given sample rate

        :param sample_rate: Sample rate used for the conversion

        :return: The frame stride in samples at the given sample rate
        """
        return self._ms_to_samples(self.frame_stride, sample_rate)


@dataclass
class DropoutConfig(KeyedClass):
    """
    Defines a dropout regularization layer

    :param rate: Probability of neurons in the input being replaced with zeros ("dropped out")
    """

    TYPE: ClassVar[str] = "dropout"

    rate: float = 0


@dataclass
class LayerNormConfig(KeyedClass):
    """
    Defines a layer normalization layer

    :param affine: If true, an affine transformation with element-wise scale and bias parameters is included
    """

    TYPE: ClassVar[str] = "layer_norm"

    affine: bool = False


@dataclass
class Glu1dConfig(KeyedClass):
    """
    Defines a one-dimensional Gated Linear Unit (GLU) layer over time

    :param out_channels: Number of filters (output channels) for the given kernel
    :param kernel: Width of the kernel
    :param stride: Stride of the kernel
    """

    TYPE: ClassVar[str] = "glu1d"

    out_channels: int
    kernel: int
    stride: int = 1


@dataclass
class MaxPoolingConfig(KeyedClass):
    """
    Defines a one-dimensional max pooling layer with a stride of one

    :param size: Size of the pooling operation
    """

    TYPE: ClassVar[str] = "max_pool"

    size: int


@dataclass
class TransformerConfig(KeyedClass):
    """
    Configures a Pre-LN transformer

    :param feedforward_neurons: Number of neurons in the feedforward sub-block
    :param heads: Number of heads in the multi-head attention sub-block
    :param activation: Activation function used between the multi-head
        attention and feedforward sub-block - either "relu" or "gelu"
    :param num_layers: Number of transformer blocks
    :param dropout_rate: Dropout rate used throughout the transformer for regularization
    :param positional_embeddings: If true, sinusoidal positional embeddings are
        added to the inputs before passing them to the transformer
    """

    TYPE: ClassVar[str] = "transformer"

    feedforward_neurons: int
    heads: int
    activation: str = typing.cast(
        str, utils.schema_field(String(default="relu", missing="relu", validate=OneOf({"relu", "gelu"})))
    )
    num_layers: int = 1
    dropout_rate: float = 0
    positional_embeddings: bool = True


class LayerSchema(KeyedOneOfSchema):
    """Schema which selects between optimizers"""

    type_field = "type"
    types = (Glu1dConfig, MaxPoolingConfig, DropoutConfig, LayerNormConfig)
    type_schemas = _choices(types)


Layer = Union[Glu1dConfig, MaxPoolingConfig, DropoutConfig, LayerNormConfig]


@dataclass
class DirectFrontendConfig(KeyedClass):
    """
    Defines a direct frontend which passes input features through as-is and optionally applies input dropout

    :param input_dropout: Amount of dropout added to the input features
    """

    TYPE: ClassVar[str] = "direct"

    input_dropout: float = 0


@dataclass
class LinearFrontendConfig(KeyedClass):
    """
    Defines a frontend with a single affine transformation and optional input dropout
    which allows the feature dimensions to be "upscaled" before passing the input to the transformer

    :param neurons: Number of neurons in the affine transformation, defines the
        input size to the transformer (`d_model`)
    :param input_dropout: Amount of dropout added to the input features
    """

    TYPE: ClassVar[str] = "linear"

    neurons: int
    input_dropout: float = 0


@dataclass
class SequentialFrontendConfig:
    """
    Defines an arbitrarily complex sequential config from a list of supported layer configurations

    :param layers: A list of supported layer configurations
    """

    layers: List[Layer] = _nested_list_field(LayerSchema)


class FrontendSchema(KeyedOneOfSchema):
    """Schema which selects between frontend architectures"""

    type_field = "architecture"
    types = (DirectFrontendConfig, LinearFrontendConfig)
    type_schemas = _choices(types)


class LossConfig(metaclass=ABCMeta):
    """
    Configures a loss function for training and validation
    """

    BLANK_OFFSET: ClassVar[int] = 0

    def get_loss(self) -> LossWrapper:
        """Retrieves the loss wrapper for a loss configuration"""
        raise NotImplementedError("No loss function defined for this class")


@dataclass
class CTCLossConfig(KeyedClass, LossConfig):
    """
    Represents regular CTC loss
    """

    TYPE: ClassVar[str] = "CTC"
    # Offset for CTC blank label
    BLANK_OFFSET: ClassVar[int] = 1

    def get_loss(self) -> CTCWrapper:
        return CTCWrapper()


@dataclass
class SequenceCrossEntropyLossConfig(KeyedClass, LossConfig):
    """
    Represents cross-entropy loss which takes the mean pooling output of the acoustic model
    instead of being applied frame-level

    :param label_smoothing: Amount of smoothing during loss computation in [0, 1], where zero disables smoothing
    """

    TYPE: ClassVar[str] = "sequence-cross-entropy"

    label_smoothing: float = dataclasses.field(default=0, metadata={"validate": Range(min=0, max=1)})

    def get_loss(self) -> SequenceCrossEntropyWrapper:
        return SequenceCrossEntropyWrapper(self.label_smoothing)


class MainLossSchema(KeyedOneOfSchema):
    """Schema which selects between loss types"""

    type_field = "type"
    types = (CTCLossConfig,)
    type_schemas = _choices(types)


class ClassifierLossSchema(KeyedOneOfSchema):
    """Schema which selects between loss types"""

    type_field = "type"
    types = (CTCLossConfig, SequenceCrossEntropyLossConfig)
    type_schemas = _choices(types)


ClassifierLossConfig = CTCLossConfig | SequenceCrossEntropyLossConfig


@dataclass
class MultiheadAttentionConfig:
    """
    A simple Multi Head Attention layer

    :param num_heads: The number of attention heads
    :param positional_embeddings: Whether to add sinusoidal positional
        embeddings to the projected attention layer input before applying multi-head attention
    """

    TYPE: ClassVar[str] = "multi-head-attention"

    num_heads: int = 1
    positional_embeddings: bool = False


class TimeLayerSchema(KeyedOneOfSchema):
    """
    Schema which selects between layer types that encode information over
    time for classifier layers in the hierarchical projection
    """

    type_field = "type"
    types = (MultiheadAttentionConfig,)
    type_schemas = _choices(types)


@dataclass
class ProjectionEntryConfig:
    """
    Configuration for a single classifier node in the hierarchical classifier and its dependencies.
    Each classifier needs at least one dependency as an input

    :param name: Name of the classifier. Should correspond to feature names
    :param dependencies: A sequence of classifier names or "OUTUT" which specifies the output of acoustic model.
        Additionally, specific layers in the acoustic model can be used as dependencies such as "OUTPUT_0" for the first hidden layer.
    :param loss: The loss applied to the classifier when training. Defaults to CTC loss
    """

    OUTPUT_DEPENDENCY: ClassVar[str] = "OUTPUT"
    OUTPUT_PATTERN: ClassVar[Pattern] = re.compile(rf"^{OUTPUT_DEPENDENCY}(?:_(\d+))?$")
    PHONEME_LAYER: ClassVar[str] = "phoneme"
    PHONE: ClassVar[str] = "phone"

    name: str
    dependencies: List[str] = dataclasses.field(default_factory=lambda: [ProjectionEntryConfig.OUTPUT_DEPENDENCY])
    time_layer: Optional[MultiheadAttentionConfig] = _nested_field(TimeLayerSchema, {"allow_none": True}, default=None)
    loss: ClassifierLossConfig = _nested_field(ClassifierLossSchema, default_factory=CTCLossConfig)


class FeatureSet(Enum):
    """
    The phonetic feature set to use
    """

    PHOIBLE = "phoible"
    PANPHON = "panphon"


class PhonemeLayerType(Enum):
    """
    Represents the type of layer used for phoneme classification
    """

    SHARED = "shared"
    PRIVATE = "private"
    ALLOPHONES = "allophones"


@dataclass
class EmbeddingCompositionConfig:
    """
    Configures an layer that uses bags of embeddings instead of linear
    projection layers for phones (with the allophone layer) and phonemes (with a shared phoneme model)

    :param embedding_size: The embedding dimension of each attribute embedding
        and, as a result, each phone or phoneme embedding vector
    """

    embedding_size: int


@dataclass
class ProjectionConfig:
    """
    Configures a hierarchical projection layer

    :param classes: A sequence of named classifiers with dependencies
    :param feature_set: The phonetic feature set for which attribute classifiers can be defined
    :param phoneme_layer: Selects the type of layer used for phoneme classification
    :param acoustic_model_dropout: Dropout to apply to (intermediate) output
        layers of the transformer before they are used as dependencies of the classifier layers
    :param dependency_blanks: If false, logits for the CTC "BLANK" class are
        removed before passing class logits through a softmax function and using
        them as the input of another classifier. Note that this only applies
        when the classifier output is used as a dependency and doesn't affect the per-class probabilities that
        can be directly accessed during training, validation or prediction
    :param allophone_l2_alpha: The alpha parameter for the allophone matrix l2
        penalty. Only used when `phoneme_layer` is set to
        :py:const:`PhonemeLayerType.PRIVATE` or :py:const:`PhonemeLayerType.ALLOPHONES`
    :param embedding_composition: Uses bags of embeddings instead of linear projection layers for phones (with the allophone layer) and phonemes (with a shared phoneme model)
    """

    classes: List[ProjectionEntryConfig]
    feature_set: FeatureSet = utils.schema_field(EnumField(FeatureSet, by_value=True), default=FeatureSet.PHOIBLE)
    phoneme_layer: PhonemeLayerType = utils.schema_field(
        EnumField(PhonemeLayerType, by_value=True), default=PhonemeLayerType.SHARED
    )
    acoustic_model_dropout: float = 0
    dependency_blanks: bool = True
    # Default value is the same as in "Universal Phone Recognition With a Multilingual Allophone Systems" by Li et al.
    allophone_l2_alpha: float = 10
    embedding_composition: Optional[EmbeddingCompositionConfig] = None

    def loss_functions(self) -> Dict[str, LossWrapper]:
        return {classifier.name: classifier.loss.get_loss() for classifier in self.classes}


FrontendConfig = DirectFrontendConfig | LinearFrontendConfig


@dataclass
class TransformerAcousticModelConfig(KeyedClass):
    """
    Defines a transformer acoustic model including architecture specific frontends

    :param transformer: Configuration for the pre-LN transformer
    :param frontend: Configuration of the frontend architecture
    :param sequential_frontend: Configuration for sequential frontend layers following the `frontend`
    :param elementwise_affine: If true, an affine transformation with
        element-wise scale and bias parameters is included in layer normalization
        immediately before, after and in the transformer acoustic model
    """

    TYPE: ClassVar[str] = "pre-ln-transformer"

    transformer: TransformerConfig
    frontend: FrontendConfig = _nested_field(FrontendSchema)
    sequential_frontend: Optional[SequentialFrontendConfig] = None
    elementwise_affine: bool = False


@dataclass
class Wav2Vec2Config(KeyedClass):
    TYPE: ClassVar[str] = "wav2vec2"


@dataclass
class UnfreezeScheduleConfig:
    """
    Declares the number of updates before unfreezing a Wav2Vec2 submodule.
    If a submodule is already unfrozen, the scheduled steps have no effect

    :param feature_encoder_steps: Nmber of steps before unfreezing the feature encoder
    :param feature_projection_steps: Nmber of steps before unfreezing the feature projection layer after the feature encoder
    :param encoder_steps: Nmber of steps before unfreezing the encoder layer after the feature projection
    """

    feature_encoder_steps: Optional[int] = None
    feature_projection_steps: Optional[int] = None
    encoder_steps: Optional[int] = None


@dataclass
class Wav2Vec2PretrainedConfig(KeyedClass):
    """
    Declares a Wav2Vec2 pre-trained model from `huggingface <https://huggingface.co/models?other=wav2vec2>`_

    :param model_id: ID of a huggingface repository containing a Wav2Vec2 pre-trained model
    :param freeze_feature_encoder: Freezes the feature encoder
    :param freeze_feature_projection: Freezes the feature projection layer after the feature encoder
    :param freeze_encoder: Freezes the encoder layer after the feature projection
    :param unfreeze_schedule: Allows unfreezing Wav2Vec2 submodules after a number of updates
    """

    TYPE: ClassVar[str] = "wav2vec2-pretrained"

    model_id: str
    freeze_feature_encoder: bool = True
    freeze_feature_projection: bool = False
    freeze_encoder: bool = False
    unfreeze_schedule: Optional[UnfreezeScheduleConfig] = None


AcousticModel = TransformerAcousticModelConfig | Wav2Vec2Config | Wav2Vec2PretrainedConfig


class AcousticModelSchema(KeyedOneOfSchema):
    """Schema which selects between acoustic models"""

    type_field = "type"
    types = (TransformerAcousticModelConfig, Wav2Vec2Config, Wav2Vec2PretrainedConfig)
    type_schemas = _choices(types)


class BatchingMode(Enum):
    """
    Represents the type of batching used which changes the unit represented by
    the provided batch size.

    With `BatchingMode.FRAMES` the `batch_size` represents the maximum number
    of frames per batch.
    With `BatchingMode.UTTERANCES` the `batch_size` represents the maximum
    number of utterances per batch. This corresponds to regular batch size.
    """

    FRAMES = "frames"
    UTTERANCES = "utterances"


@dataclass
class Architecture:
    """
    Configuration options for the acoustic model architecture

    :param batch_size: The batch size used during training. Its unit depends on the value of `batching_mode`
    :param projection: Configuration of the hierarchical projection layer
    :param acoustic_model: Configuration of the acoustic model
    :param optimizer: Configuration for the optimizer
    :param loss: Configuration for the loss function
    :param early_stopping_patience: The number of steps with no improvement of validation loss before stopping training early
    :param batching_mode: Represents the type of batching used which changes the unit represented by the provided batch size
    :param language_oversampling_factor: Factor to use for oversampling from lower resource languages when generting batches for training
    :param seed: The random seed used for all otherwise non-deterministic operations
    :param maximum_iterations: Maximum number of epochs before the training is terminated
        If `None` is given, it acts as infinity
    :param clip_norm: Norm used for gradient clipping
    :param lr_schedule: Configuration for a learning rate schedule
    :param lr_warmup: Configuration for learning rate warmup
    :param accumulation_factor: Factor for dividing batches into mini-batches
        with gradient accumulation to allow for larger batch sizes with less memory
    :param step_size: Number of batch updates before performing validation for very large
        datasets. If `None` is given, validation is only performed once after each epoch
    :param mixed_precision: Enables efficient mixed precision training with auto scaling
    """

    batch_size: int
    projection: ProjectionConfig
    acoustic_model: AcousticModel = _nested_field(AcousticModelSchema)
    optimizer: Optimizer = _nested_field(OptimizerSchema)
    loss: CTCLossConfig = _nested_field(MainLossSchema)
    early_stopping_patience: Optional[int] = None
    batching_mode: BatchingMode = utils.schema_field(
        EnumField(BatchingMode, by_value=True), default=BatchingMode.FRAMES
    )
    language_oversampling_factor: Optional[float] = None
    seed: Optional[int] = None
    maximum_iterations: Optional[int] = None
    clip_norm: Optional[float] = None
    lr_schedule: Optional[LrSchedulerConfig] = _nested_field(LrSchedulerSchema, {"allow_none": True}, default=None)
    accumulation_factor: int = 1
    step_size: Optional[int] = None
    mixed_precision: bool = False


class FeatureType(Enum):
    """
    Type of acoustic features used by the network
    """

    MFCC = "MFCC"
    FILTERBANKS = "Filterbanks"
    RAW = "raw"


@dataclass
class Preprocessing:
    """
    Confguration for audio preprocessing

    :param window: Window for which the given `feature_type` is computed
    :param feature_type: Type of acoustic features to compute during preprocessing
    :param resample: Sampling rate to resample audio to before computing acoustic features
    :param num_filters: Number of filters for the given `feature_type`
    """

    window: Window
    feature_type: FeatureType = utils.schema_field(EnumField(FeatureType))
    resample: Optional[int] = 16_000  # 16kHz
    num_filters: int = 40


@dataclass
class DataConfig:
    """
    Configuration for the training data

    :param languages: List of languages to use for training from a multilingual corpus
    :param validation_limits: Limits the amount of utterances used as
        validation data, either a map from language codes to language specific limits,
        a single limit for every language or `None` for no limit
    :param only_primary_script: Keeps only transcriptions that consist of a
        single script that corresponds to the primary script of each language
    """

    languages: List[str]
    validation_limits: Dict[str, int] | int | None = None
    only_primary_script: bool = False


@dataclass
class ProfilingConfig:
    """
    Configuration for performance and memory profiling of training code

    :param active_steps: Number of batch update steps for which the profiler will collect samples
    :param flame_graph_path_cpu: Target path for the cpu traces
    :param flame_graph_path_gpu: Target path for the gpu traces
    :param tensorboard_dir: Path to a tensorboard log directory that profiling information will be stored in.
        When this isn't `None`, flame_graph_path_* variables will be ignored.
    :param profile_memory: Whether to profile memory consumption
    :param record_shapes: Whether to record tensors shapes during pytorch computations
    :param repeat: Number of epochs for which to repeat profiling
    """

    active_steps: int
    flame_graph_path_gpu: Optional[str] = None
    flame_graph_path_cpu: Optional[str] = None
    tensorboard_dir: Optional[str] = None
    profile_memory: bool = False
    record_shapes: bool = False
    repeat: int = 1


ConfigCls = TypeVar("ConfigCls", bound="Config")


@marshmallow_dataclass.dataclass
class Config:
    """
    Configuration for preprocessing, training and the architecture of an acoustic model

    :param nn: Configuration for the neural network architecture
    :param preprocessing: Configuration for the preprocessing of acoustic features
    :param data: Configuration for the training data
    :param profiling: Configuration for performance and memory profiling of training code
    """

    Schema: ClassVar[Type[Schema]]

    nn: Architecture
    preprocessing: Preprocessing
    data: DataConfig
    profiling: Optional[ProfilingConfig] = None

    @classmethod
    def load(cls: Type[ConfigCls], mapping: Mapping[str, Any]) -> ConfigCls:
        return cls.Schema().load(mapping)  # type: ignore

    @classmethod
    def from_toml(cls: Type[ConfigCls], path: str) -> ConfigCls:
        return cls.Schema().load(toml.load(path))  # type: ignore

    def dump(self) -> Dict[str, Any]:
        return self.Schema().dump(self)  # type: ignore

    def dumps_toml(self) -> str:
        return toml.dumps(self.dump())

    def dumps(self) -> str:
        return self.Schema().dumps(self)



from collections.abc import Callable, Sequence
from dataclasses import dataclass
import os
import re
from typing import Dict, List, Literal, Union, overload
import typing

import zarr
from zarr import Group
import numpy as np
import torch
from torch import Tensor

from allophant import utils
from allophant.language_codes import LanguageCodeMap
from allophant.datasets.mozilla_common_voice import CommonVoiceCorpus, CommonVoiceSplits
from allophant.datasets.speech_corpus import MultilingualSplits
from allophant.datasets.ucla_phonetic_corpus import UCLAPhoneticCorpus
from allophant import utils
from allophant.utils import AnyPath


def _map_languages(directory: AnyPath, languages: Sequence[str]) -> List[str]:
    language_map = LanguageCodeMap(CommonVoiceCorpus.read_meta_from(directory).language_codes())
    return [language_map[code] for code in languages]


@overload
def load_corpus(
    path: AnyPath,
    dataset_type: Literal["ucla-phonetic"],
    resample: int | None = None,
    languages: Sequence[str] | None = None,
    validation_limits: None = None,
    only_primary_script: Literal[False] = False,
    progress_bar: bool = False,
) -> MultilingualSplits[UCLAPhoneticCorpus]: ...


@overload
def load_corpus(
    path: AnyPath,
    dataset_type: Literal["common-voice"],
    resample: int | None = None,
    languages: Sequence[str] | None = None,
    validation_limits: Dict[str, int] | int | None = None,
    only_primary_script: bool = False,
    progress_bar: bool = False,
) -> CommonVoiceSplits: ...


def load_corpus(
    path: AnyPath,
    dataset_type: Union[Literal["common-voice"], Literal["ucla-phonetic"]],
    resample: int | None = None,
    languages: Sequence[str] | None = None,
    validation_limits: Dict[str, int] | int | None = None,
    only_primary_script: bool = False,
    progress_bar: bool = False,
) -> MultilingualSplits:
    if dataset_type == "common-voice":
        # Loads the full corpus if a directory is given or a single split from a transcription file otherwise
        if os.path.isdir(path):
            if languages is not None:
                languages = _map_languages(path, languages)
            return CommonVoiceSplits.from_file(
                path,
                resample,
                languages,
                validation_limits=validation_limits,
                only_primary_script=only_primary_script,
                progress_bar=progress_bar,
            )

        data_directory = os.path.dirname(path)
        if languages is not None:
            languages = _map_languages(data_directory, languages)

        if (
            split_match := re.match(
                utils.format_parse_pattern(CommonVoiceCorpus.TRANSCRIPTION_PATTERN), os.path.basename(path)
            )
        ) is None:
            raise ValueError(f"Path is not a valid transcription path: {path!r}")

        (split,) = split_match.groups()
        if split not in CommonVoiceSplits.SPLIT_NAMES:
            raise ValueError(f"{split} is not a valid Split, must be one of {CommonVoiceSplits.SPLIT_NAMES}")

        # Manually cast since literal inference is not yet supported
        split = typing.cast(Literal["train", "dev", "test"], split)
        return CommonVoiceSplits.single(
            CommonVoiceCorpus.from_file(
                data_directory,
                split,
                resample,
                languages,
                only_primary_script=only_primary_script,
                progress_bar=progress_bar,
            ),
            split,
        )

    elif dataset_type == "ucla-phonetic":
        # Loaded as a test split
        return MultilingualSplits.single(UCLAPhoneticCorpus.load(path, resample, languages, progress_bar))

    raise ValueError(f"Corpus of type {dataset_type} is not supported")


@dataclass
class PreprocessedSplitData:
    lengths: Tensor
    features: List[np.ndarray] | None = None


def _reshape_to_features(feature_dimension: int) -> Callable[[np.ndarray], np.ndarray]:
    def reshape(array: np.ndarray) -> np.ndarray:
        return array.reshape((-1, feature_dimension))

    return reshape


def _load_features_and_lengths(
    split_data: Group, languages: Sequence[str], feature_size: int = 1
) -> PreprocessedSplitData:
    features = []
    lengths = []
    for language in languages:
        language_data = split_data[language]
        features.extend(map(_reshape_to_features(feature_size), language_data["features"]))
        lengths.append(language_data["lengths"])

    return PreprocessedSplitData(torch.from_numpy(np.concatenate(lengths)), features)


def _lengths_for_split(split_data: Group, languages: Sequence[str]) -> PreprocessedSplitData:
    return PreprocessedSplitData(
        torch.from_numpy(np.concatenate([split_data[language]["lengths"] for language in languages]))  # type: ignore
    )


def preprocessed_features_or_lengths(
    data_path: str, split_languages: Dict[str, Sequence[str]], lengths_only: bool = True
) -> Dict[str, PreprocessedSplitData]:
    feature_data = zarr.open_group(data_path, "r")
    if feature_data is None:
        raise IOError("Feature data couldn't be loaded")

    feature_size = None if lengths_only else feature_data.attrs["feature_size"]
    split_groups = []
    split_data = {}
    for split in split_languages:
        split_group = feature_data[split]
        assert isinstance(split_group, Group), f"Invalid feature format - split {split} is not a group"
        if feature_size is None:
            split_data[split] = _lengths_for_split(split_group, split_languages[split])
        else:
            split_data[split] = _load_features_and_lengths(split_group, split_languages[split], feature_size)
        split_groups.append(split_group)

    return split_data



from collections.abc import Set
from importlib.metadata import version
from typing import (
    Callable,
    Iterable,
    Iterator,
    Optional,
    Dict,
    Any,
    Tuple,
    Union,
    TypeVar,
    Type,
    ClassVar,
    List,
    Protocol,
)
from dataclasses import dataclass, field
from contextlib import contextmanager
import itertools
import math
from datetime import timedelta
import timeit
import contextlib
import typing
import gc
import os

import marshmallow_dataclass
import torch
from torch import profiler
from torch.profiler import ProfilerActivity
from torch import Tensor
from torch import nn
from torch import cuda
from torch.cuda.amp import autocast_mode
from torch.cuda.amp.grad_scaler import GradScaler
from torch.utils.tensorboard.writer import SummaryWriter
from marshmallow import Schema
from marshmallow.fields import Raw
from tqdm import tqdm
import numpy as np
import transformers

from allophant import utils
from allophant.attribute_graph import AttributeGraph, AttributeGraphField, AttributeNode
from allophant.dataset_processing import BatchType, TranscribedDataset, SamplesProcessor
from allophant.datasets.speech_corpus import AudioInfo, MultilingualCorpus, MultilingualSplits, SplitMetaData
from allophant.phonetic_features import PhoneticAttributeIndexer, PhoneticIndexerState
from allophant.utils import OnlineMean, PathOrFile, PathOrFileBinary
from allophant.network.acoustic_model import Allophant, Predictions, UnfreezeSchedule
from allophant.config import (
    Config,
    OptimizerWrapper,
    ProfilingConfig,
    ProjectionEntryConfig,
    WarmupInfo,
    Wav2Vec2PretrainedConfig,
)
from allophant.batching import Batch, Batcher, LabeledBatch
from allophant.loss_functions import LossWrapper


@dataclass
class TrainingStatus:
    """
    Named tuple representing the training status of an `Estimator` indicating
    whether training should stop and whether there was an improvement over the
    previous best epoch
    """

    stop: bool
    improvement: bool


class StoppingCriterion(Protocol):
    """Protocol for a stopping criterion for `Estimator` training"""

    def status(self, validation_losses: float) -> TrainingStatus:
        """
        Checks whether training should stop and whether there was an
        improvement over the previous best epoch based on validation losses

        :param validation_losses: Losses on the validation set on the current epoch

        :return: A `TrainingStatus` for the current epoch
        """
        raise NotImplementedError


class ImprovementTrackingCriterion(StoppingCriterion):
    """
    Basic :py:class:`StoppingCriterion` that never stops but reports validation loss improvements
    """

    def __init__(self):
        """Initializes the criterion"""
        self._minimum_losses = math.inf

    def status(self, validation_losses: float) -> TrainingStatus:
        if validation_losses < self._minimum_losses:
            self._minimum_losses = validation_losses
            return TrainingStatus(False, True)

        return TrainingStatus(False, False)


class EarlyStopping(StoppingCriterion):
    """
    Stops training after a set number of epochs without improvement on validation set loss
    """

    def __init__(self, patience: int):
        """
        Initializes early stopping with the given patience.

        :param patience: Number of epochs without improvement on validation set loss before stopping training
        """
        self.patience = patience
        self._minimum_losses = math.inf
        self._epochs_without_improvement = 0

    def status(self, validation_losses: float) -> TrainingStatus:
        if validation_losses < self._minimum_losses:
            self._minimum_losses = validation_losses
            self._epochs_without_improvement = 0
            # Continue training after improvement
            return TrainingStatus(False, True)

        self._epochs_without_improvement += 1
        # Decide whether to continue after no improvement
        return TrainingStatus(self._epochs_without_improvement == self.patience, False)


M = TypeVar("M", bound=nn.Module)


@contextmanager
def evaluation(model: M) -> Iterator[M]:
    """
    Context manager which temporarily sets a model into evaluation mode

    :param model: The target model

    :return: The `model` in evaluation mode
    """
    original_mode = model.training
    model.eval()
    yield model
    model.train(original_mode)


@dataclass
class EpochPosition:
    epoch: int = 0
    global_step: int = 0
    step: int = 0

    def next_step(self) -> None:
        self.global_step += 1
        self.step += 1

    def __str__(self) -> str:
        return f"Epoch {self.epoch}" + (
            "" if self.step is None else f", Step {self.step}" + f" | Global Step: {self.global_step}"
        )


@dataclass
class EpochStatistics:
    """Stores statistics about loss and accuracy for an epoch"""

    epoch: EpochPosition
    training_loss: float
    validation_loss: float
    training_seconds: float = 0
    validation_seconds: float = 0

    def __str__(self) -> str:
        epoch_delta = timedelta(seconds=self.training_seconds + self.validation_seconds)
        base_info = (
            f"{self.epoch}"
            f" | Training loss: {self.training_loss:.4f}"
            f" | Validation loss: {self.validation_loss:.4f}"
            f" | Δt: {epoch_delta}"
        )
        return base_info


@dataclass
class OptimizationStates:
    optimizer: Dict[str, Tensor] = utils.schema_field(Raw(default=None, missing=None))
    grad_scaler: Optional[Dict[str, Any]] = None


CheckpointCls = TypeVar("CheckpointCls", bound="Checkpoint")


@marshmallow_dataclass.add_schema
@dataclass
class Checkpoint:
    """
    Holds data for an `Estimator` checkpoint and functions for serialization and deserialization
    """

    Schema: ClassVar[Type[Schema]]

    config: Config
    allophant_version: str
    feature_size: int
    sample_rate: int
    attribute_graph: AttributeGraph = field(metadata={"marshmallow_field": AttributeGraphField()})
    epoch: EpochPosition
    phonetic_indexer_state: PhoneticIndexerState
    dataset_meta_data: List[SplitMetaData]
    model_state: Dict[str, Tensor] = utils.schema_field(Raw())
    additional: Optional[Dict[str, Any]] = utils.schema_field(Raw(default=None, missing=None))
    history: List[Tuple[TrainingStatus, EpochStatistics]] = field(default_factory=list)
    optimization_states: Optional[OptimizationStates] = None

    def save(self, file: PathOrFileBinary) -> None:
        """
        Saves the checkpoint to a path or file object

        :param file: path or file object the checkpoint is saved to
        """
        torch.save(self.Schema().dump(self), file)

    @classmethod
    def restore(
        cls: Type[CheckpointCls], file: PathOrFile, device: torch.device | str | None = None, **kwargs
    ) -> CheckpointCls:
        """
        Restores a checkpoint from a huggingface *model id*, a *path* or a
        *file object*. Additional keyword arguments are passed to :py:func:`transformers.utils.cached_file`.

        :param file: huggingface *model id*, *path* or *file object* the checkpoint is read from
        :param device: Target device for model weights, specify "cuda" for loading weights on the GPU

        :return: A `Checkpoint` which was read from the `file`
        """
        # Try loading a cached file from huggingface if a string path is given but does not exist locally
        if isinstance(file, str) and not os.path.isfile(file):
            resolved = transformers.utils.cached_file(file, "allophant.pt", **kwargs)
            if resolved is None:
                raise FileNotFoundError(f"No checkpoint found at {file!r}")
            file = resolved

        return cls.Schema().load(torch.load(file, map_location=device, weights_only=True))  # type: ignore


def split_batch_size(batch_size: int, accumulation_factor: int) -> int:
    batch_size, remainder = divmod(batch_size, accumulation_factor)
    if remainder > 0:
        raise ValueError(f"Batch size {batch_size} is not divisble by the accumulation factor {accumulation_factor}")
    return batch_size


@dataclass
class TrainDevLengths:
    train: Tensor
    development: Tensor


@dataclass
class TrainDevFeatures:
    train: List[np.ndarray]
    dev: List[np.ndarray]


def _generate_attribute_graph_data(
    phonetic_indexer: PhoneticAttributeIndexer,
    config: Config,
) -> Iterator[AttributeNode]:
    for phoneme_attribute in config.nn.projection.classes:
        yield AttributeNode(
            phoneme_attribute.name,
            phonetic_indexer.size(phoneme_attribute.name),
            phoneme_attribute.time_layer,
            phoneme_attribute.dependencies,
        )


DatasetManagerCls = TypeVar("DatasetManagerCls", bound="DatasetManager")


class DatasetManager:
    def __init__(
        self,
        config: Config,
        data_splits: MultilingualSplits[MultilingualCorpus],
        samples_processor: SamplesProcessor,
        data_lengths: TrainDevLengths | None = None,
        data_features: TrainDevFeatures | None = None,
        data_workers: int | None = 0,
        perform_validation: bool = True,
    ):
        nn_config = config.nn
        self._index_start_offset = nn_config.loss.BLANK_OFFSET
        batch_size = split_batch_size(nn_config.batch_size, nn_config.accumulation_factor)
        # Only the custom Transformer Acoustic Model requires packed batches while Wav2Vec2 uses custom processing
        self.training_batcher = Batcher(
            batch_size, nn_config.batching_mode, nn_config.language_oversampling_factor, data_workers
        )
        self.validation_batcher = Batcher(batch_size, nn_config.batching_mode, data_workers=data_workers)
        # Reuse the global seed specifically for the batcher to ensure consistent data sampling for model comparisons
        self._batching_seed = config.nn.seed

        self._data_splits = data_splits
        self._data_workers = data_workers
        self._perform_validation = perform_validation
        self._sample_processor = samples_processor
        if data_lengths is None:
            self._training_lengths = self._development_lengths = None
        else:
            self._training_lengths = data_lengths.train
            self._development_lengths = data_lengths.development

        self._training_data = self._process_dataset(
            data_splits.train, None if data_features is None else data_features.train
        )
        self._validation_data = self._process_dataset(
            data_splits.dev,
            None if data_features is None else data_features.dev,
            # Map phonemes that might not appear in the training data
            {
                language_id: set(inventory)
                for language_id, inventory in data_splits.train.language_id_inventories().inventories.items()
            },
        )

        self._resampled = config.preprocessing.resample
        self._training_set_size = len(self._data_splits.train)
        self._validation_set_size = len(self._data_splits.dev)
        self._audio_info = self._data_splits.audio_info()
        self._sample_rate = self._audio_info.sample_rate

    def _process_dataset(
        self,
        data_split: MultilingualCorpus,
        features: List[np.ndarray] | None = None,
        phoneme_mapper: Dict[int, Set[str]] | None = None,
    ) -> TranscribedDataset:
        return TranscribedDataset(
            BatchType.INDEXED, data_split, self._sample_processor, self._index_start_offset, features, phoneme_mapper
        )

    def training_batches(self, shuffle: bool = True) -> Iterator[LabeledBatch]:
        return self.training_batcher.batches(self._training_data, self._training_lengths, shuffle, self._batching_seed)

    def development_batches(self, shuffle: bool = False) -> Iterator[LabeledBatch]:
        return self.validation_batcher.batches(
            self._validation_data, self._development_lengths, shuffle, self._batching_seed
        )

    @property
    def splits(self) -> MultilingualSplits:
        return self._data_splits

    @property
    def training_set_size(self) -> int:
        return self._training_set_size

    @property
    def validation_set_size(self) -> int:
        return self._validation_set_size

    @property
    def feature_size(self) -> int:
        return self._sample_processor.feature_size

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @property
    def audio_info(self) -> AudioInfo:
        return self._audio_info

    def indexer_state(self) -> PhoneticIndexerState:
        return self._sample_processor.indexer_state()

    def attribute_graph(self, config: Config) -> AttributeGraph:
        return AttributeGraph(_generate_attribute_graph_data(self._sample_processor.attribute_indexer, config))

    @classmethod
    def from_config(
        cls: Type[DatasetManagerCls],
        config: Config,
        data_splits: MultilingualSplits,
        attribute_indexer: PhoneticAttributeIndexer,
        data_lengths: Optional[TrainDevLengths] = None,
        data_features: Optional[TrainDevFeatures] = None,
        data_workers: Optional[int] = 0,
        perform_validation: bool = True,
    ) -> DatasetManagerCls:
        sample_processor = SamplesProcessor.from_config(
            config,
            data_splits.audio_info().sample_rate,
            attribute_indexer,
        )
        return cls(config, data_splits, sample_processor, data_lengths, data_features, data_workers, perform_validation)


def profiler_trace_handler(profiling_config: ProfilingConfig) -> Callable[[Any], None]:
    if profiling_config.tensorboard_dir is not None:
        return profiler.tensorboard_trace_handler(profiling_config.tensorboard_dir)

    flame_graph_path_cpu = profiling_config.flame_graph_path_cpu
    flame_graph_path_gpu = profiling_config.flame_graph_path_gpu

    def trace_handler(profiling_scheduler) -> None:
        if flame_graph_path_cpu is not None:
            profiling_scheduler.export_stacks(flame_graph_path_cpu, "self_cpu_time_total")
        if flame_graph_path_gpu is not None:
            profiling_scheduler.export_stacks(flame_graph_path_gpu, "self_cuda_time_total")

    return trace_handler


EstimatorCls = TypeVar("EstimatorCls", bound="Estimator")


class TrainingParameters:
    def __init__(
        self,
        stopping_criterion: Optional[StoppingCriterion] = None,
        max_iterations: Optional[int] = None,
        validate: bool = True,
        progress: bool = True,
        summary_writer: Optional[SummaryWriter] = None,
    ) -> None:
        self.max_iterations = max_iterations
        self.validate = validate
        self.progress = progress
        self.stopping_criterion = stopping_criterion
        self.summary_writer = summary_writer


@dataclass
class TrainingProgressStatistics:
    batch_mean: OnlineMean = field(default_factory=OnlineMean)
    batch_sizes: List[int] = field(default_factory=list)
    running_mean_training_losses: float = 0
    training_losses: float = 0
    training_lengths: float = 0

    def update_mean(self, batch_training_lengths: int) -> float:
        self.training_lengths += batch_training_lengths
        self.running_mean_training_losses = self.training_losses / self.training_lengths
        return self.running_mean_training_losses


class CategoryBatchStatistics:
    def __init__(self, classes: List[str]):
        self.losses = {category: 0.0 for category in classes}
        self.lengths = typing.cast(Dict[str, int], self.losses.copy())

    def add(self, name: str, losses: float, length: int) -> None:
        self.losses[name] += losses
        self.lengths[name] += length

    def sum_lengths(self) -> int:
        return sum(self.lengths.values())

    def mean_losses(self) -> Dict[str, float]:
        return {name: losses / self.lengths[name] for name, losses in self.losses.items()}


def _create_training_progress_bar(
    training_set_size: int, step_size: Optional[int] = None, progress: bool = True
) -> tqdm:
    return tqdm(
        total=training_set_size if step_size is None else step_size,
        unit="utterances" if step_size is None else "update",
        desc="Training",
        disable=not progress,
    )


def _clear_cache() -> None:
    gc.collect()
    cuda.empty_cache()


class TrainingRun:
    def __init__(
        self,
        estimator: "Estimator",
        dataset_manager: DatasetManager,
        optimizer: OptimizerWrapper,
        training_parameters: TrainingParameters | None = None,
    ) -> None:
        self._estimator = estimator
        self._config = estimator.config
        self._model = estimator.model
        self._loss_functions = estimator.loss_functions

        self._dataset_manager = dataset_manager
        self._optimizer = optimizer
        self._parameters = TrainingParameters() if training_parameters is None else training_parameters

        if self._parameters.stopping_criterion is None:
            patience = self._config.nn.early_stopping_patience
            self._parameters.stopping_criterion = (
                ImprovementTrackingCriterion() if patience is None else EarlyStopping(patience)
            )

        self._stopping_criterion = self._parameters.stopping_criterion
        self._available_memory = cuda.get_device_properties(0).total_memory
        self._max_batch_elements = 0
        acoustic_model_config = self._config.nn.acoustic_model
        if (
            isinstance(acoustic_model_config, Wav2Vec2PretrainedConfig)
            and acoustic_model_config.unfreeze_schedule is not None
        ):
            self._unfreeze_schedule = UnfreezeSchedule.from_config(acoustic_model_config.unfreeze_schedule)
        else:
            self._unfreeze_schedule = None

    def _log_training_step(
        self,
        summary_writer: SummaryWriter,
        global_step: int,
        batch_size: int,
        progress_statistics: TrainingProgressStatistics,
        category_statistics: CategoryBatchStatistics,
    ):
        summary_writer.add_scalars("Training/Loss/Classifiers", category_statistics.mean_losses(), global_step)
        summary_writer.add_scalar("Training/Loss/Mean", progress_statistics.running_mean_training_losses, global_step)
        # Compute gradient norms to log
        total_norm = 0
        for parameter in self._model.parameters():
            gradient = parameter.grad
            if gradient is None or not parameter.requires_grad:
                continue
            param_norm = gradient.detach().data.norm()
            total_norm += param_norm.item() ** 2

        summary_writer.add_scalar("Training/GradientNorm", math.sqrt(total_norm), global_step)
        summary_writer.add_scalar("Training/LearningRate", self._optimizer.current_learning_rate(), global_step)
        summary_writer.add_scalars(
            "Training/GPUMemory",
            {
                "cached": cuda.memory_reserved() / self._available_memory * 100,
                "actual": cuda.memory_allocated() / self._available_memory * 100,
                "max_cached": cuda.max_memory_reserved() / self._available_memory * 100,
                "max_actual": cuda.max_memory_allocated() / self._available_memory * 100,
            },
            global_step,
        )
        summary_writer.add_scalar("Training/Batch/Size", batch_size, global_step)
        summary_writer.add_scalar("Training/Batch/MaximumElementsWithPadding", self._max_batch_elements, global_step)
        summary_writer.flush()

    def _log_validation_step(
        self,
        summary_writer: SummaryWriter,
        global_step: int,
        validation_losses: float,
        category_statistics: CategoryBatchStatistics,
    ) -> None:
        summary_writer.add_scalars("Validation/Loss/Classifiers", category_statistics.mean_losses(), global_step)
        summary_writer.add_scalar("Validation/Loss/Mean", validation_losses, global_step)
        summary_writer.flush()

    def _finish_validation(
        self,
        epoch: EpochPosition,
        training_delta: float,
        validation_start_time: float,
        average_training_losses: float,
        average_validation_losses: float,
    ) -> Tuple[TrainingStatus, EpochStatistics]:
        validation_delta = (timeit.default_timer() - validation_start_time) if self._parameters.validate else 0
        epoch_statistics = EpochStatistics(
            epoch,
            average_training_losses,
            average_validation_losses,
            training_delta,
            validation_delta,
        )

        # Get and record state and statistics of the training step or epoch
        training_state = (self._stopping_criterion.status(epoch_statistics.validation_loss), epoch_statistics)
        self._estimator.history.append(training_state)
        return training_state

    @torch.inference_mode()
    def _end_step(
        self,
        epoch: EpochPosition,
        training_device: torch.device,
        training_start_time: float,
        average_training_losses: float,
        use_mixed_precision: bool = False,
    ) -> Tuple[TrainingStatus, EpochStatistics]:
        training_delta = timeit.default_timer() - training_start_time
        total_validation_loss = 0
        sum_validation_lengths = 0

        validation_start_time = timeit.default_timer()
        if not self._parameters.validate:
            return self._finish_validation(
                epoch,
                training_delta,
                validation_start_time,
                average_training_losses,
                # Reports training losses as validation losses if validation is disabled
                # for compatibility with stopping schedules
                average_training_losses,
            )

        with (
            evaluation(self._model),
            tqdm(
                total=self._dataset_manager.validation_set_size,
                unit="utterances",
                desc="Validation",
                leave=False,
                disable=not self._parameters.progress,
                position=0,
            ) as progress_bar,
        ):
            category_statistics = CategoryBatchStatistics(self._model.classes)
            validation_mean = 0
            batch_mean = OnlineMean()

            for batch in self._dataset_manager.development_batches():
                batch_size = len(batch)
                batch_mean += batch_size
                with autocast_mode.autocast() if use_mixed_precision else contextlib.nullcontext():
                    batch = batch.to(training_device)
                    predictions = self._model(batch)
                    validation_loss = 0
                    [batch_labels] = batch.attribute_indices
                    [label_lengths] = batch.label_lengths
                    # Silently remove phone node since it is never explicitly trained
                    predictions.outputs.pop(ProjectionEntryConfig.PHONE, None)

                    for name, output in predictions.outputs.items():
                        category_lengths = label_lengths[batch.label_length_indices[name]]
                        labels = batch_labels[name]

                        category_loss = self._loss_functions[name](
                            output,
                            labels,
                            predictions.lengths,
                            category_lengths,
                        ).item()

                        total_attribute_label_lengths = int(category_lengths.sum().item())
                        category_statistics.add(name, category_loss, total_attribute_label_lengths)

                        validation_loss += category_loss
                        sum_validation_lengths += total_attribute_label_lengths

                total_validation_loss += float(validation_loss)
                validation_mean = total_validation_loss / sum_validation_lengths

                progress_bar.update(len(batch))
                progress_bar.set_postfix({"average batch size": float(batch_mean), "loss": validation_mean})

        summary_writer = self._parameters.summary_writer
        if summary_writer is not None:
            self._log_validation_step(summary_writer, epoch.global_step, validation_mean, category_statistics)

        return self._finish_validation(
            epoch,
            training_delta,
            validation_start_time,
            average_training_losses,
            total_validation_loss / sum_validation_lengths,
        )

    def _create_epoch_counter(self, start: int = 0) -> Iterable[int]:
        if self._parameters.max_iterations is None:
            return itertools.count(start)
        return range(start, self._parameters.max_iterations + 1)

    def _scaled_backward(self, losses: Tensor) -> None:
        if self._scaler is not None:
            typing.cast(Tensor, self._scaler.scale(losses)).backward()
        else:
            losses.backward()

    def _training_batch_accumulation(
        self,
        batches: List[LabeledBatch],
        progress_statistics: TrainingProgressStatistics,
        progress_bar: tqdm,
        use_mixed_precision: bool,
        training_device: torch.device,
    ) -> CategoryBatchStatistics:
        allophone_l2_alpha = self._config.nn.projection.allophone_l2_alpha
        step_size = self._config.nn.step_size

        category_statistics = CategoryBatchStatistics(self._model.classes)
        # Number of accumulated batches before draining them gradually from the list
        batch_count = len(batches)

        while len(batches) > 0:
            batch = batches.pop().to(training_device)
            batch_size = len(batch)
            label_length_indices = batch.label_length_indices

            with autocast_mode.autocast() if use_mixed_precision else contextlib.nullcontext():
                predictions = self._model(batch)
                output_lengths = predictions.lengths
                mini_batch_training_loss = torch.tensor(0, dtype=torch.float32, device=output_lengths.device)

                [batch_labels] = batch.attribute_indices
                [label_lengths] = batch.label_lengths

                # Remove batch early
                del batch

                # Silently remove phone node since it is never explicitly trained
                predictions.outputs.pop(ProjectionEntryConfig.PHONE, None)

                for name, output in predictions.outputs.items():
                    category_lengths = label_lengths[label_length_indices[name]]
                    labels = batch_labels[name]

                    category_loss = self._loss_functions[name](
                        output,
                        labels,
                        output_lengths,
                        category_lengths,
                    )

                    # Record losses for each category and on average
                    category_statistics.add(name, category_loss.item(), int(category_lengths.sum().item()))
                    mini_batch_training_loss += category_loss

            # Accumulate gradients for each mini batch after computing the mean over losses for each frame
            mini_batch_mean_training_losses = mini_batch_training_loss / category_statistics.sum_lengths()
            self._scaled_backward(mini_batch_mean_training_losses)
            # Running loss for epoch statistics
            progress_statistics.training_losses += mini_batch_training_loss.item()

            progress_bar.update(batch_size if step_size is None else 1 / batch_count)
            progress_statistics.batch_mean += batch_size
            progress_statistics.batch_sizes.append(batch_size)

            progress_bar.set_postfix(
                {
                    "average batch size": float(progress_statistics.batch_mean),
                    "loss": float(progress_statistics.running_mean_training_losses),
                    "lr": self._optimizer.current_learning_rate(),
                }
            )

            # Adds l2 penalty after accumulating losses over the whole batch
            with autocast_mode.autocast() if use_mixed_precision else contextlib.nullcontext():
                l2_penalty = self._model.l2_penalty()
            if l2_penalty is not None:
                self._scaled_backward(allophone_l2_alpha * l2_penalty)

        return category_statistics

    def _training_step(
        self,
        batches: List[LabeledBatch],
        progress_statistics: TrainingProgressStatistics,
        progress_bar: tqdm,
        use_mixed_precision: bool,
        training_device: torch.device,
    ) -> CategoryBatchStatistics:
        category_statistics = self._training_batch_accumulation(
            batches,
            progress_statistics,
            progress_bar,
            use_mixed_precision,
            training_device,
        )

        clip_norm = self._config.nn.clip_norm

        if clip_norm is not None:
            # Unscales gradients before gradient clipping as in https://pytorch.org/docs/stable/notes/amp_examples.html#gradient-clipping
            if self._scaler is not None:
                self._scaler.unscale_(self._optimizer.optimizer)
            nn.utils.clip_grad.clip_grad_norm_(self._model.parameters(), clip_norm)

        # Perform optimization step and scaler update with mixed precision after accumulating backward passes
        if self._scaler is not None:
            self._scaler.step(self._optimizer)
            self._scaler.update()
        else:
            self._optimizer.step()

        if self._unfreeze_schedule is not None:
            self._unfreeze_schedule.step(self._model.acoustic_model)

        progress_statistics.update_mean(category_statistics.sum_lengths())

        return category_statistics

    def __iter__(self) -> Iterator[Tuple[TrainingStatus, EpochStatistics]]:
        training_device = next(self._model.parameters()).device

        # Enables mixed precision if it's both enabled and training is performed on a GPU
        use_cuda = training_device.type == "cuda"
        use_mixed_precision = self._config.nn.mixed_precision and use_cuda
        self._scaler = GradScaler() if use_mixed_precision else None

        step_size = self._config.nn.step_size
        accumulation_factor = self._config.nn.accumulation_factor

        # Epoch starts at 1 with epoch 0 indicating no training
        epoch_position = EpochPosition()

        profiling_config = self._config.profiling
        trace_handler = None if profiling_config is None else profiler_trace_handler(profiling_config)

        progress_statistics = TrainingProgressStatistics()
        step_update = 0
        training_start_time = timeit.default_timer()

        progress_bar = _create_training_progress_bar(
            self._dataset_manager.training_set_size, step_size, self._parameters.progress
        )
        for epoch in self._create_epoch_counter(epoch_position.epoch):
            # Set epoch based on the counter
            epoch_position.epoch = epoch
            self._estimator.epoch = epoch_position

            with (
                profiler.profile(
                    activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA] if use_cuda else [ProfilerActivity.CPU],
                    schedule=profiler.schedule(wait=1, warmup=2, active=profiling_config.active_steps),
                    record_shapes=profiling_config.record_shapes,
                    profile_memory=profiling_config.profile_memory,
                    # Required by flamegraph
                    with_stack=profiling_config.flame_graph_path_cpu is not None
                    or profiling_config.flame_graph_path_gpu is not None,
                    on_trace_ready=trace_handler,
                )
                if profiling_config is not None
                else contextlib.nullcontext()
            ) as profiling_scheduler:
                if step_size is not None:
                    # Reset at the start of an epoch only if epoch based training is used
                    progress_statistics = TrainingProgressStatistics()

                # NOTE: Requires higher ulimit or copy.deepcopy in some cases due to file descriptor limit
                batch_generator = self._dataset_manager.training_batches(shuffle=True)
                while batches := list(itertools.islice(batch_generator, accumulation_factor)):
                    epoch_position.next_step()
                    step_update += 1

                    batch_size = 0
                    for batch in batches:
                        batch_size += batch.audio_features.shape[0]
                        batch_elements_count = batch.audio_features.numel()
                        if batch_elements_count > self._max_batch_elements:
                            self._max_batch_elements = batch_elements_count

                    category_statistics = self._training_step(
                        batches, progress_statistics, progress_bar, use_mixed_precision, training_device
                    )

                    # Clear cache every 100 steps
                    if not epoch_position.global_step % 100:
                        _clear_cache()

                    summary_writer = self._parameters.summary_writer
                    if summary_writer is not None:
                        self._log_training_step(
                            summary_writer,
                            epoch_position.global_step,
                            batch_size,
                            progress_statistics,
                            category_statistics,
                        )

                    # Gradient zeroing is performed after potential logging in tensorboard
                    self._optimizer.optimizer.zero_grad(set_to_none=True)

                    if profiling_scheduler is not None:
                        profiling_scheduler.step()

                    if step_size is not None and step_size == step_update:
                        # Reset step
                        step_update = 0
                        # Clear output/move line upwards
                        progress_bar.reset()
                        status, epoch_statistics = self._end_step(
                            epoch_position,
                            training_device,
                            training_start_time,
                            progress_statistics.running_mean_training_losses,
                            use_mixed_precision,
                        )
                        # Reset progress statistics
                        progress_statistics = TrainingProgressStatistics()

                        training_start_time = timeit.default_timer()
                        # Validation loss is the same as training loss if trained with validate = False
                        yield status, epoch_statistics
                        if status.stop:
                            return

                # Perform validation at an epoch step if step-level validation is disabled
                if step_size is None:
                    # Clear output/move line upwards
                    progress_bar.reset()
                    _clear_cache()
                    status, epoch_statistics = self._end_step(
                        epoch_position,
                        training_device,
                        training_start_time,
                        progress_statistics.running_mean_training_losses,
                        use_mixed_precision,
                    )
                    _clear_cache()

                    # Reset progress statistics
                    progress_statistics = TrainingProgressStatistics()

                    training_start_time = timeit.default_timer()
                    # Validation loss is the same as training loss if trained with validate = False
                    yield status, epoch_statistics
                    if status.stop:
                        return

        progress_bar.close()


@dataclass
class Estimator:
    """
    Wrapper around an acoustic model, supporting training and
    prediction as well as saving and restoring checkpoints
    """

    config: Config
    feature_size: int
    sample_rate: int
    attribute_graph: AttributeGraph
    model: Allophant
    loss_functions: Dict[str, LossWrapper]
    history: List[Tuple[TrainingStatus, EpochStatistics]] = field(default_factory=list)
    epoch: EpochPosition = field(default_factory=EpochPosition)
    dataset_meta_data: List[SplitMetaData] = field(default_factory=list)

    def __post_init__(self):
        self._scaler = None

    @classmethod
    def from_config(
        cls: Type[EstimatorCls],
        config: Config,
        feature_size: int,
        sample_rate: int,
        attribute_graph: AttributeGraph,
        attribute_indexer: PhoneticAttributeIndexer | None = None,
        device: torch.device | str = "cuda",
        load_pretrained_weights: bool = True,
    ) -> EstimatorCls:
        model = Allophant.from_config(
            config.nn,
            feature_size,
            sample_rate,
            attribute_graph,
            attribute_indexer,
            load_pretrained_weights,
        )
        model = model.to(device)

        return cls(
            config,
            feature_size,
            sample_rate,
            attribute_graph,
            model,
            config.nn.projection.loss_functions(),
        )

    def create_optimizer(self) -> OptimizerWrapper:
        wrapper = self.config.nn.optimizer.get_optimizer(self.model.parameters(), WarmupInfo(self.model.d_model))
        wrapper.add_schedulers(self.config.nn.lr_schedule)

        return wrapper

    def _prepare_optimizer(
        self, optimizer: OptimizerWrapper, optimization_states: Optional[OptimizationStates] = None
    ) -> OptimizerWrapper:
        # Add plateau schedule if necessary before restoring the optimizer state if required
        optimizer.add_schedulers(self.config.nn.lr_schedule)

        # Restore states for the optimizer and the gradient scaler
        if optimization_states is not None:
            optimizer.load_state_dict(optimization_states.optimizer)
            if self._scaler is not None and optimization_states.grad_scaler is not None:
                self._scaler.load_state_dict(optimization_states.grad_scaler)

        return optimizer

    def train(
        self,
        dataset_manager: DatasetManager,
        optimizer: OptimizerWrapper,
        stopping_criterion: Optional[StoppingCriterion] = None,
        max_iterations: Optional[int] = None,
        validate: bool = True,
        progress: bool = True,
        optimization_states: Optional[OptimizationStates] = None,
        summary_writer: Optional[SummaryWriter] = None,
    ) -> TrainingRun:
        """
        Generator which performs one step or epoch of training and validation each iteration.

        :param dataset_manager: Dataset manager containing a data reader, a phoneme preprocessing function and batchers
        :param optimizer: `OptimizerWrapper` used for training
        :param stopping_criterion: A stopping criterion which supplies the `TrainingStatus` for each iteration.
            If `TrainingStatus.stop` is `True`, the training generator will stop
        :param max_iterations: Maximum number of epochs. If not given training will only stop based on the `stopping_criterion`
        :param validate: When set to `False` validation is disabled and the
            stopping criterion will be applied to training losses instead

        :return: Yields a tuple of the `TrainingStatus` and `EpochStatistics` for each step or epoch
        """
        # Track the last dataset meta data used for training
        self.dataset_meta_data.append(dataset_manager.splits.meta_data())

        return TrainingRun(
            self,
            dataset_manager,
            self._prepare_optimizer(optimizer, optimization_states),
            TrainingParameters(stopping_criterion, max_iterations, validate, progress, summary_writer),
        )

    @torch.inference_mode()
    def predict(
        self, batch: Batch, target_feature_indices: Tensor | None = None, log_probabilities: bool = True
    ) -> Predictions:
        with evaluation(self.model):
            predictions = self.model(batch, target_feature_indices, predict=True)
            if log_probabilities:
                return Predictions(
                    {name: self.model.log_probabilities(logits) for name, logits in predictions.outputs.items()},
                    predictions.lengths,
                )
            return predictions

    def map_allophones(self, phone_logits: Tensor, language_ids: Tensor) -> Tensor:
        return self.model.map_allophones(phone_logits, language_ids)

    def save(
        self,
        file: PathOrFileBinary,
        optimizer_state: Dict[str, Tensor],
        phonetic_indexer_state: PhoneticIndexerState,
        additional_parameters: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Saves a checkpoint containing model parameters and configuration to a file.
        Checkpoints can be restored using `Estimator.restore`

        :param file: Path or file object the current checkpoint will be saved to
        :param optimizer_state: The `state_dict` of the optimizer used during training
        :param additional_parameters: Dictionary of additional parameters which
            can later be retrieved from the saved `Checkpoint`
        """
        Checkpoint(
            config=self.config,
            feature_size=self.feature_size,
            sample_rate=self.sample_rate,
            attribute_graph=self.attribute_graph,
            epoch=self.epoch,
            phonetic_indexer_state=phonetic_indexer_state,
            dataset_meta_data=self.dataset_meta_data,
            allophant_version=version(__package__),
            model_state=self.model.state_dict(),
            optimization_states=OptimizationStates(
                optimizer_state,
                self._scaler.state_dict() if self._scaler is not None else None,
            ),
            additional=additional_parameters,
            history=self.history,
        ).save(file)

    @classmethod
    def restore(
        cls: Type[EstimatorCls], checkpoint: Union[Checkpoint, PathOrFile], device: torch.device | str = "cuda", **kwargs
    ) -> Tuple[EstimatorCls, PhoneticAttributeIndexer]:
        """
        Restores an `Estimator` from a checkpoint previously saved using
        `Estimator.save`. Additional keyword arguments are passed to
        :py:func:`transformers.utils.cached_file`.

        :param checkpoint_or_file: A `Checkpoint` which was previously saved
            using `Estimator.save` from which the model state and config are
            restored. It can either be:
            - a `Checkpoint` instance
            - a huggingface *model id*
            - a local *path* or *file object*
        :param device: Target device for model weights, specify "cuda" for
            loading weights on the GPU

        :return: A tuple of the `Estimator` restored from the given checkpoint and the phonetic attributes used by the model
        """
        # Restore parameters
        if not isinstance(checkpoint, Checkpoint):
            checkpoint = Checkpoint.restore(checkpoint, device=device, **kwargs)

        phoneme_indexer = PhoneticAttributeIndexer.from_config(
            checkpoint.config, state_dict=checkpoint.phonetic_indexer_state
        )

        estimator = cls.from_config(
            checkpoint.config,
            checkpoint.feature_size,
            checkpoint.sample_rate,
            checkpoint.attribute_graph,
            phoneme_indexer,
            device,
            load_pretrained_weights=False,
        )
        estimator.model.load_state_dict(checkpoint.model_state)
        estimator.epoch = checkpoint.epoch
        estimator.history = checkpoint.history

        return estimator, phoneme_indexer



from collections.abc import Callable, KeysView, Iterable, Iterator, Sequence
import contextlib
import csv
from dataclasses import dataclass, field
from datetime import date
from importlib.metadata import version
import itertools
import json
import os
from os import path
from typing import Any, ClassVar, Dict, Generic, List, Literal, Optional, Tuple, Type, TypeVar, Union, overload

from marshmallow import EXCLUDE, Schema
from marshmallow.fields import Raw
import marshmallow_dataclass
import requests
import msgpack
from msgpack import ExtType
from mashumaro.mixins.msgpack import DataClassMessagePackMixin
from tqdm.contrib.logging import logging_redirect_tqdm
from tqdm import tqdm

from allophant import utils
import allophant
from allophant.config import FeatureSet
from allophant.datasets.phonemes import PhonemeTranscriber, PhonemeTranscription, PhonemeSource, TaggedTranscription
from allophant.datasets.speech_corpus import (
    LanguageData,
    LanguageInfo,
    MultilingualCorpus,
    MultilingualSplits,
    PhoneticallyTranscribedUtterance,
)
from allophant import csv_validator
from allophant.csv_validator import CsvSchema
from allophant.phoneme_segmentation import SEGMENTATION_LOGGER
from allophant.phonetic_features import PhoneticAttributeIndexer
from allophant.utils import AnyPath, CamelCasingSchema, MarshmallowDataclassLoadMixin


def _custom_data_key(key: str) -> Any:
    return field(metadata={"data_key": key})


def _optional_zero_field() -> Any:
    return field(default=0, metadata={"allow_none": True})


@dataclass
class CategoryStatistics:
    accent: Dict[str, float]
    age: Dict[str, float]
    gender: Dict[str, float]


@dataclass
class SplitSizes:
    dev: int = 0
    invalidated: int = 0
    other: int = 0
    reported: int = 0
    test: int = 0
    train: int = 0
    validated: int = 0


@dataclass
class LocaleMeta:
    buckets: SplitSizes
    clips: int
    splits: CategoryStatistics
    users: int
    size: int
    checksum: Optional[str]
    avg_duration_secs: float = _optional_zero_field()
    valid_duration_secs: float = _optional_zero_field()
    total_hrs: float = _optional_zero_field()
    valid_hrs: float = _optional_zero_field()
    duration: int = _optional_zero_field()
    reported_sentences: int = 0


@marshmallow_dataclass.add_schema(base_schema=CamelCasingSchema)
@dataclass
class ReleaseMeta(MarshmallowDataclassLoadMixin):
    date: date
    name: str
    multilingual: bool
    locales: Dict[str, LocaleMeta]
    total_duration: int
    total_valid_duration_secs: int
    total_hrs: int
    total_valid_hrs: int
    bundle_url_template: str = _custom_data_key("bundleURLTemplate")
    Schema: ClassVar[Type[Schema]]

    def __getitem__(self, locale: str) -> LocaleMeta:
        return self.locales[locale]

    def __iter__(self) -> Iterator[Tuple[str, LocaleMeta]]:
        yield from self.locales.items()

    def language_codes(self) -> KeysView[str]:
        return self.locales.keys()

    @classmethod
    def download(cls, version: str):
        return cls.from_json(download_release_meta(version))

    class Meta:
        unknown = EXCLUDE


LANGUAGE_META_URL_TEMPLATE = "https://commonvoice.mozilla.org/dist/releases/cv-corpus-{}.json"


def download_release_meta(version: str) -> Any:
    return requests.get(LANGUAGE_META_URL_TEMPLATE.format(version)).json()


@dataclass
class RawTranscription(DataClassMessagePackMixin):
    raw_sentence: str
    utterance_id: str
    client_id: str
    age: Optional[str]
    gender: Optional[str]
    accents: Optional[str]

    _EXT_ID = 2


@dataclass
class Transcription(RawTranscription, DataClassMessagePackMixin):
    phonemes: PhonemeTranscription

    Schema: ClassVar[Type[Schema]]
    _EXT_ID = 4


_TRANSCRIPTION_TYPES = (Transcription, RawTranscription)


def transcription_ext_encoder(obj: Any) -> ExtType:
    for transcription_type in _TRANSCRIPTION_TYPES:
        if isinstance(obj, transcription_type):
            return ExtType(transcription_type._EXT_ID, obj.to_msgpack())
    raise TypeError(f"Received object of unknown type: {type(obj)}")


def transcription_ext_decoder(code: int, data: bytes):
    for transcription_type in _TRANSCRIPTION_TYPES:
        if code == transcription_type._EXT_ID:
            return transcription_type.from_msgpack(data)
    return ExtType(code, data)


CommonVoiceRawData = LanguageData[LanguageInfo, RawTranscription]
CommonVoicePhonemeData = LanguageData[LanguageInfo, Transcription]


T = TypeVar("T", bound=Union[RawTranscription, Transcription])


@marshmallow_dataclass.add_schema
@dataclass
class CommonVoiceCorpusMeta(MarshmallowDataclassLoadMixin):
    # Uses Union and Optional instead of Python 3.10 union syntax as a workaround: https://github.com/lidatong/dataclasses-json/issues/349
    corpus_name: str
    phoneme_sources: Optional[List[str]] = None
    feature_set: Union[FeatureSet, str, None] = None
    limits: Union[Dict[str, int], int, None] = None
    utterance_counts: Dict[str, int] = field(default_factory=dict)
    package_version: str = version(allophant.__package__)


@marshmallow_dataclass.add_schema
@dataclass
class CommonVoiceCorpusSplit(Generic[T], MarshmallowDataclassLoadMixin):
    meta: CommonVoiceCorpusMeta
    inventory_mappings: Dict[str, List[Dict[str, List[str]]]]
    transcriptions: Dict[str, List[T]] = utils.schema_field(Raw())


class CommonVoiceCorpus(MultilingualCorpus, Generic[T]):
    UTTERANCE_TYPE = PhoneticallyTranscribedUtterance

    CORPUS_PATH_FILE = ".corpus_path"
    META_FILE = "meta.json"
    TRANSCRIPTION_PATTERN = "{split}_transcriptions.bin"
    INVENTORY_PATTERN = "{split}_inventories.json"

    # Sample rates identified in CommonVoice version 10.0
    _SAMPLE_RATES = [8000, 16000, 24000, 32000, 44100, 48000]
    _AUDIO_DIRECTORY = "clips"

    def __init__(
        self,
        base_directory: AnyPath,
        languages: Iterable[LanguageData[LanguageInfo, T]],
        meta_data: CommonVoiceCorpusMeta,
        resample: int | None = None,
        phoneme_sources: Sequence[PhonemeSource] | None = None,
        limits: Dict[str, int] | int | None = None,
        data_directory: AnyPath | None = None,
    ) -> None:
        super().__init__(base_directory, languages, self._AUDIO_DIRECTORY, "mp3", limits, resample, phoneme_sources)
        # Set and update meta data after corpus has been fully loaded
        meta_data.utterance_counts = {
            language_subset.info.code: len(language_subset.transcribed_samples) for language_subset in self
        }
        self._original_meta_data = meta_data
        self._meta_data = meta_data.to_json()
        self._data_directory = data_directory

        self._limits = limits

    def __str__(self) -> str:
        return f"Mozilla Common Voice Corpus containing {len(self._languages)} languages with {self._num_utterances} utterances"

    @property
    def data_directory(self) -> AnyPath | None:
        return self._data_directory

    @classmethod
    def write_corpus_path(cls, directory: AnyPath, corpus_directory: AnyPath) -> None:
        with open(path.join(directory, cls.CORPUS_PATH_FILE), "w", encoding="utf-8") as file:
            file.write(str(corpus_directory))

    @classmethod
    def get_corpus_path(cls, directory: AnyPath) -> AnyPath:
        path_file_path = path.join(directory, cls.CORPUS_PATH_FILE)
        if not path.isfile(path_file_path):
            return directory
        with open(path_file_path, "r", encoding="utf-8") as file:
            return file.read().strip("\r\n")

    @classmethod
    def _read_meta_directly(cls, direct_directory: AnyPath) -> ReleaseMeta:
        return ReleaseMeta.load(path.join(direct_directory, cls.META_FILE))

    @classmethod
    def read_meta_from(cls, directory: AnyPath) -> ReleaseMeta:
        return cls._read_meta_directly(cls.get_corpus_path(directory))

    def read_meta(self) -> ReleaseMeta:
        return self._read_meta_directly(self._base_directory)

    @classmethod
    def load_split(
        cls,
        directory: AnyPath,
        split: Literal["train"] | Literal["dev"] | Literal["test"],
        g2p_engine: PhonemeTranscriber | None = None,
        feature_set: FeatureSet | str | None = None,
        include_single_upvote_other: bool = False,
        batch_size: int = 1,
        language_codes: Sequence[str] | None = None,
        map_to_allophoible: bool = False,
        limits: Dict[str, int] | int | None = None,
        use_progress_bar: bool = False,
        mapping_threshold: int | None = None,
    ):
        # Ignore files in the top level directory like meta.json and transcriptions.json
        language_codes = (
            [sub_dir for sub_dir in os.listdir(directory) if path.isdir(path.join(directory, sub_dir))]
            if language_codes is None
            else language_codes
        )

        inventory_indexer = (
            PhoneticAttributeIndexer(
                FeatureSet.PHOIBLE,
                language_inventories=language_codes,
                allophones_from_allophoible=True,
            )
            if map_to_allophoible
            else None
        )

        if g2p_engine is not None:
            for language in language_codes:
                if not g2p_engine.supports(language):
                    raise ValueError(f"Language {language!r} not supported by all G2P engines")
                if not g2p_engine.supports_tokenization(language):
                    raise ValueError(f"No available tokenizer model for {language!r}")

        if use_progress_bar:
            # Reset logger to get warnings for each invocation
            SEGMENTATION_LOGGER.reset()

        phoneme_sources = None if g2p_engine is None else g2p_engine.phoneme_sources

        with logging_redirect_tqdm() if use_progress_bar else contextlib.nullcontext():
            return cls(
                directory,
                _load_common_voice_splits(
                    directory,
                    language_codes,
                    [split, _OTHER_SPLIT] if include_single_upvote_other else [split],
                    g2p_engine,
                    batch_size,
                    inventory_indexer,
                    limits,
                    use_progress_bar,
                    mapping_threshold,
                ),
                CommonVoiceCorpusMeta(
                    path.basename(path.normpath(directory)),
                    None if phoneme_sources is None else [source.value for source in phoneme_sources],
                    feature_set,
                    limits,
                ),
                phoneme_sources=phoneme_sources,
            )

    def transcriptions_to_json(self) -> Dict[str, Dict[str, Any]]:
        transcriptions = {}
        inventory_mappings = {}
        for language in self.languages:
            subset = self.monolingual_subset(language)
            transcriptions[language] = subset.transcribed_samples
            inventory_mappings[language] = subset.info.phoneme_mappings

        return CommonVoiceCorpusSplit(
            self._original_meta_data,
            inventory_mappings,
            transcriptions,
        ).to_json()

    def save(self, corpus_directory: AnyPath, split: str, output_directory: Optional[AnyPath] = None) -> None:
        if output_directory is None:
            output_directory = corpus_directory
        else:
            os.mkdir(output_directory)
            self.write_corpus_path(output_directory, corpus_directory)

        with open(
            path.join(output_directory, self.TRANSCRIPTION_PATTERN.format(split=split)), "wb"
        ) as transcription_file:
            msgpack.pack(self.transcriptions_to_json(), transcription_file, default=transcription_ext_encoder)
        with open(
            path.join(output_directory, self.INVENTORY_PATTERN.format(split=split)), "w", encoding="utf-8"
        ) as inventory_file:
            json.dump(
                {language: self.monolingual_subset(language).info.phoneme_inventory for language in self.languages},
                inventory_file,
            )

    @classmethod
    def load(
        cls,
        transcriptions: Dict[str, Any],
        inventories: Dict[str, List[str]],
        data_directory: AnyPath,
        base_directory: AnyPath,
        resample: Optional[int] = None,
        languages: Optional[Sequence[str]] = None,
        limits: Dict[str, int] | int | None = None,
        only_primary_script: bool = False,
    ):
        split = CommonVoiceCorpusSplit.from_json(transcriptions)
        transcribed_utterances = split.transcriptions
        inventory_mappings = split.inventory_mappings

        if languages is None:
            allowed_languages = None
        else:
            # NOTE: Assumes language codes are premapped
            allowed_languages = set(languages)
            missing_languages = allowed_languages - transcribed_utterances.keys()
            if missing_languages:
                raise KeyError(
                    f"Languages are missing form the locally stored Common Voice dataset: {missing_languages}"
                )

        return cls(
            base_directory,
            (
                LanguageData(
                    LanguageInfo(language, inventories[language], inventory_mappings[language]),
                    # Keep only transcriptions that consist of a single script that corresponds to the primary script of the language if `only_primary_script` is enabled
                    (
                        [
                            transcription
                            for transcription in transcriptions
                            if transcription.phonemes.only_primary_script()
                        ]
                        if only_primary_script
                        else transcriptions
                    ),
                )
                for language, transcriptions in transcribed_utterances.items()
                if allowed_languages is None or language in allowed_languages
            ),
            CommonVoiceCorpusMeta(
                split.meta.corpus_name,
                split.meta.phoneme_sources,
                split.meta.feature_set,
                # Either store the current limits in the metadata or limits set during initial data preprocessing
                (limits or split.meta.limits),
                split.meta.utterance_counts,
                split.meta.package_version,
            ),
            resample,
            (
                None
                if split.meta.phoneme_sources is None
                else [PhonemeSource(variant) for variant in split.meta.phoneme_sources]
            ),
            limits,
            data_directory,
        )

    @classmethod
    def from_file(
        cls,
        data_directory: AnyPath,
        split: Union[Literal["train"], Literal["dev"], Literal["test"]],
        resample: Optional[int] = None,
        languages: Optional[Sequence[str]] = None,
        corpus_directory: Optional[AnyPath] = None,
        limits: Dict[str, int] | int | None = None,
        only_primary_script: bool = False,
        progress_bar: bool = False,
    ):
        if corpus_directory is None:
            corpus_directory = cls.get_corpus_path(data_directory)
        with (
            utils.file_and_path_wrapper(
                path.join(data_directory, cls.TRANSCRIPTION_PATTERN.format(split=split)), "rb"
            ) as file,
            (
                tqdm.wrapattr(file, "read", total=path.getsize(file.name), desc=f"Loading transcriptions for {split}")
                if progress_bar
                else file
            ) as wrapped,
        ):
            transcriptions = msgpack.unpack(wrapped, ext_hook=transcription_ext_decoder)
        with utils.file_and_path_wrapper(
            path.join(data_directory, cls.INVENTORY_PATTERN.format(split=split)), "r", encoding="utf-8"
        ) as file:
            inventories = json.load(file)

        return cls.load(
            transcriptions,
            inventories,
            data_directory,
            corpus_directory,
            resample,
            languages,
            limits,
            only_primary_script,
        )

    def load_inventories_for(self, split: str) -> Dict[str, List[str]]:
        with utils.file_and_path_wrapper(
            path.join(self._data_directory or self._base_directory, self.INVENTORY_PATTERN.format(split=split)),
            "r",
            encoding="utf-8",
        ) as file:
            return json.load(file)


@dataclass
class CommonVoiceSplits(Generic[T], MultilingualSplits[CommonVoiceCorpus[T]]):
    def transcriptions_to_json(self) -> Dict[str, Any]:
        return {
            "train": self.train.transcriptions_to_json(),
            "dev": self.dev.transcriptions_to_json(),
            "test": self.test.transcriptions_to_json(),
        }

    def save(self, corpus_directory: AnyPath, output_directory: Optional[AnyPath] = None) -> None:
        if output_directory is None:
            output_directory = corpus_directory
        else:
            os.mkdir(output_directory)
            CommonVoiceCorpus.write_corpus_path(output_directory, corpus_directory)

        for split in self.SPLIT_NAMES:
            getattr(self, split).save(output_directory, split)

    @classmethod
    def load(
        cls,
        transcriptions: Dict[str, Any],
        inventories: Dict[str, List[str]],
        data_directory: AnyPath,
        base_directory: AnyPath,
        resample: int | None = None,
        languages: Sequence[str] | None = None,
        validation_limits: Dict[str, int] | int | None = None,
        only_primary_script: bool = False,
    ):
        train, dev, test = cls.SPLIT_NAMES
        return cls(
            CommonVoiceCorpus.load(
                transcriptions[train],
                inventories,
                data_directory,
                base_directory,
                resample,
                languages,
                only_primary_script=only_primary_script,
            ),
            CommonVoiceCorpus.load(
                transcriptions[dev],
                inventories,
                data_directory,
                base_directory,
                resample,
                languages,
                validation_limits,
                only_primary_script,
            ),
            CommonVoiceCorpus.load(
                transcriptions[test],
                inventories,
                data_directory,
                base_directory,
                resample,
                languages,
                only_primary_script=only_primary_script,
            ),
        )

    @classmethod
    def from_file(
        cls,
        directory: AnyPath,
        resample: Optional[int] = None,
        languages: Optional[Sequence[str]] = None,
        corpus_directory: Optional[AnyPath] = None,
        validation_limits: Dict[str, int] | int | None = None,
        only_primary_script: bool = False,
        progress_bar: bool = False,
    ):
        if corpus_directory is None:
            corpus_directory = CommonVoiceCorpus.get_corpus_path(directory)

        train, dev, test = cls.SPLIT_NAMES
        return cls(
            CommonVoiceCorpus.from_file(
                directory,
                train,
                resample,
                languages,
                corpus_directory,
                only_primary_script=only_primary_script,
                progress_bar=progress_bar,
            ),
            CommonVoiceCorpus.from_file(
                directory,
                dev,
                resample,
                languages,
                corpus_directory,
                validation_limits,
                only_primary_script,
                progress_bar,
            ),
            CommonVoiceCorpus.from_file(
                directory,
                test,
                resample,
                languages,
                corpus_directory,
                only_primary_script=only_primary_script,
                progress_bar=progress_bar,
            ),
        )


_OTHER_SPLIT = "other"


@dataclass
class _CommonVoiceEntry:
    client_id: str
    path: str
    sentence: str
    up_votes: int
    down_votes: int
    age: Optional[str]
    gender: Optional[str]
    accents: Optional[str]
    locale: str
    segment: Optional[str]

    @classmethod
    def schema(cls):
        return csv_validator.make_schema(_CommonVoiceEntry)


def _extract_entries(
    language_path: AnyPath,
    splits: Sequence[str],
    split_filters: Sequence[Callable[[_CommonVoiceEntry], bool]],
    schema: CsvSchema[_CommonVoiceEntry],
    limit: Optional[int] = None,
    use_progress_bar: bool = False,
) -> Iterator[_CommonVoiceEntry]:
    processed_utterances = 0

    for split, split_filter in zip(splits, split_filters):
        with open(path.join(language_path, split + ".tsv"), "r", encoding="utf-8") as file:
            if use_progress_bar:
                # Get the number of all entries without the header
                entry_count = sum(1 for _ in file) - 1
                file.seek(0)
                file = tqdm(
                    file,
                    # Get the number of remaining entries until the limit or the total number of entries if no limit is given
                    total=entry_count if limit is None else min(limit - processed_utterances, entry_count),
                    unit=" utterances",
                    desc=f"Processing {path.basename(language_path)}/{split}",
                    position=1,
                    leave=False,
                )
                processed_utterances += entry_count

            tsv_file = csv.reader(file, delimiter="\t", quoting=csv.QUOTE_NONE)
            # Skip header
            next(tsv_file, None)
            for i, (line, _) in enumerate(zip(tsv_file, utils.limit_indices(limit))):
                try:
                    entry = schema.convert_line(line)
                    if not split_filter(entry):
                        continue

                    yield entry
                except csv.Error as error:
                    # Schema validation errors with more details
                    raise csv.Error(f"In line {i}: {line}") from error


@overload
def _load_common_voice_splits(
    directory: AnyPath,
    languages: Sequence[str],
    splits: Sequence[str],
    g2p_engine: PhonemeTranscriber,
    batch_size: int,
    inventory_indexer: Optional[PhoneticAttributeIndexer] = None,
    limits: Optional[Union[int, Dict[str, int]]] = None,
    use_progress_bar: bool = False,
    mapping_threshold: int | None = None,
) -> Iterator[CommonVoicePhonemeData]: ...


@overload
def _load_common_voice_splits(
    directory: AnyPath,
    languages: Sequence[str],
    splits: Sequence[str],
    g2p_engine: None,
    batch_size: int,
    inventory_indexer: Optional[PhoneticAttributeIndexer] = None,
    limits: Optional[Union[int, Dict[str, int]]] = None,
    use_progress_bar: bool = False,
    mapping_threshold: int | None = None,
) -> Iterator[CommonVoiceRawData]: ...


def _load_common_voice_splits(
    directory: AnyPath,
    languages: Sequence[str],
    splits: Sequence[str],
    g2p_engine: Optional[PhonemeTranscriber] = None,
    batch_size: int = 1,
    inventory_indexer: Optional[PhoneticAttributeIndexer] = None,
    limits: Optional[Union[int, Dict[str, int]]] = None,
    use_progress_bar: bool = False,
    mapping_threshold: int | None = None,
) -> Iterator[Union[CommonVoiceRawData, CommonVoicePhonemeData]]:
    def true(_: _CommonVoiceEntry) -> Literal[True]:
        return True

    def positive_score(entry: _CommonVoiceEntry) -> bool:
        return (entry.up_votes - entry.down_votes) > 1

    schema = _CommonVoiceEntry.schema()

    split_filters = [positive_score if split == _OTHER_SPLIT else true for split in splits]
    for language in tqdm(languages, position=0, unit=" languages") if use_progress_bar else languages:
        SEGMENTATION_LOGGER.context_language(language)
        language_path = path.join(directory, language)
        phoneme_inventories = [set() for _ in range(g2p_engine.num_engines if g2p_engine is not None else 0)]
        transcribed = []
        language_entries = _extract_entries(
            language_path,
            splits,
            split_filters,
            schema,
            utils.global_or_local_limit(limits, language),
            use_progress_bar,
        )

        if g2p_engine is None:
            for entry in language_entries:
                transcribed.append(
                    RawTranscription(
                        entry.sentence,
                        path.splitext(entry.path)[0],
                        entry.client_id,
                        entry.age,
                        entry.gender,
                        entry.accents,
                    )
                )
        else:
            for entries, transcription_batches in g2p_engine.extractor(language).auto_batch_g2p_transcribe(
                language_entries, batch_size
            ):
                for entry, transcriptions in zip(entries, transcription_batches):
                    for inventory, transcription in zip(
                        phoneme_inventories, transcriptions.flattened_primary_transcriptions()
                    ):
                        inventory.update(transcription)
                    transcribed.append(
                        Transcription(
                            entry.sentence,
                            path.splitext(entry.path)[0],
                            entry.client_id,
                            entry.age,
                            entry.gender,
                            entry.accents,
                            transcriptions,
                        )
                    )

        # Sort phoneme inventories for consistency
        phoneme_inventories = [sorted(inventory) for inventory in phoneme_inventories]
        # Create phoneme mapping and remap transcriptions for each engine
        if inventory_indexer is None:
            phoneme_inventory = sorted(phoneme for inventory in phoneme_inventories for phoneme in inventory)
            inventory_mappings = []
        else:
            inventory_mappings = inventory_indexer.map_language_inventory(
                phoneme_inventories, language, distance_threshold=mapping_threshold
            )

            # Only store phonemes in the inventory that occur after remapping
            phoneme_inventory = sorted(
                set(
                    mapped_phoneme
                    for mapping, inventory in zip(inventory_mappings, phoneme_inventories)
                    for phoneme in inventory
                    for mapped_phoneme in mapping[phoneme]
                )
            )

            # Remap phonemes from all transcriptions
            for transcription in transcribed:
                transcription.phonemes.phonemes = [
                    [
                        (
                            TaggedTranscription(
                                list(
                                    itertools.chain.from_iterable(mapping[phoneme] for phoneme in segment.transcription)
                                ),
                                segment.language,
                            )
                            if segment.language is None
                            else segment
                        )
                        for segment in phoneme_transcriptions
                    ]
                    for mapping, phoneme_transcriptions in zip(inventory_mappings, transcription.phonemes.phonemes)
                ]

        yield LanguageData(
            LanguageInfo(
                language,
                # Merge original inventories from multiple engines together
                phoneme_inventory,
                inventory_mappings,
            ),
            transcribed,
        )


@overload
def load_common_voice(
    directory: AnyPath,
    g2p_engine: PhonemeTranscriber,
    feature_set: FeatureSet | str,
    include_single_upvote_other: bool,
    batch_size: int,
    languages: Optional[Sequence[str]],
    map_to_allophoible: bool = False,
    training_limits: Optional[Union[int, Dict[str, int]]] = None,
    use_progress_bar: bool = False,
    mapping_threshold: int | None = None,
) -> CommonVoiceSplits[Transcription]: ...


@overload
def load_common_voice(
    directory: AnyPath,
    g2p_engine: None,
    feature_set: None,
    include_single_upvote_other: bool,
    batch_size: int,
    languages: Optional[Sequence[str]],
    map_to_allophoible: bool = False,
    training_limits: Optional[Union[int, Dict[str, int]]] = None,
    use_progress_bar: bool = False,
    mapping_threshold: int | None = None,
) -> CommonVoiceSplits[RawTranscription]: ...


def load_common_voice(
    directory: AnyPath,
    g2p_engine: Optional[PhonemeTranscriber] = None,
    feature_set: FeatureSet | str | None = None,
    include_single_upvote_other: bool = False,
    batch_size: int = 1,
    languages: Optional[Sequence[str]] = None,
    map_to_allophoible: bool = False,
    training_limits: Optional[Union[int, Dict[str, int]]] = None,
    use_progress_bar: bool = False,
    mapping_threshold: int | None = None,
) -> CommonVoiceSplits:
    if use_progress_bar:
        # Share logger memory over all three slits
        SEGMENTATION_LOGGER.suppress_resets(3)

    train_split = CommonVoiceCorpus.load_split(
        directory,
        "train",
        g2p_engine,
        feature_set,
        include_single_upvote_other,
        batch_size,
        languages,
        map_to_allophoible,
        training_limits,
        use_progress_bar,
        mapping_threshold,
    )
    return CommonVoiceSplits(
        train_split,
        CommonVoiceCorpus.load_split(
            directory,
            "dev",
            g2p_engine,
            feature_set,
            map_to_allophoible=map_to_allophoible,
            batch_size=batch_size,
            language_codes=train_split.languages,
            use_progress_bar=use_progress_bar,
            mapping_threshold=mapping_threshold,
        ),
        # Don't remap test phonemes regardless of settings for more freedom at test time
        CommonVoiceCorpus.load_split(
            directory,
            "test",
            g2p_engine,
            feature_set,
            map_to_allophoible=False,
            batch_size=batch_size,
            language_codes=train_split.languages,
            use_progress_bar=use_progress_bar,
        ),
    )
